import ctypes

from ...exceptions import CanInitializationError, CanOperationError

try:
    # try to load libraries from the system default paths
    _csi = ctypes.windll.LoadLibrary("dll-csiBind")
    _oci = ctypes.windll.LoadLibrary("dll-ocdProxy")
except FileNotFoundError:
    # try to load libraries with hardcoded paths
    if ctypes.sizeof(ctypes.c_voidp) == 4:
        # 32 bit
        path = "C:/Program Files (x86)/ETAS/BOA_V2/Bin/Win32/Dll/Framework/"
    elif ctypes.sizeof(ctypes.c_voidp) == 8:
        # 64 bit
        path = "C:/Program Files/ETAS/BOA_V2/Bin/x64/Dll/Framework/"
    _csi = ctypes.windll.LoadLibrary(path + "dll-csiBind")
    _oci = ctypes.windll.LoadLibrary(path + "dll-ocdProxy")


# define helper functions to use with errcheck


def errcheck_init(result, func, _arguments):
    # unfortunately, we can't use OCI_GetError here
    # because we don't always have a handle to use
    # text = ctypes.create_string_buffer(500)
    # OCI_GetError(self.ctrl, ec, text, 500)
    if result != 0x0:
        raise CanInitializationError(f"{func.__name__} failed with error 0x{result:X}")
    return result


def errcheck_oper(result, func, _arguments):
    if result != 0x0:
        raise CanOperationError(f"{func.__name__} failed with error 0x{result:X}")
    return result


# Common (BOA)

BOA_ResultCode = ctypes.c_uint32
BOA_Handle = ctypes.c_int32
BOA_Time = ctypes.c_int64

BOA_NO_VALUE = -1
BOA_NO_HANDLE = BOA_Handle(BOA_NO_VALUE)
BOA_NO_TIME = BOA_Time(BOA_NO_VALUE)


class BOA_UuidBin(ctypes.Structure):
    _fields_ = [("data", ctypes.c_uint8 * 16)]


class BOA_Version(ctypes.Structure):
    _fields_ = [
        ("majorVersion", ctypes.c_uint8),
        ("minorVersion", ctypes.c_uint8),
        ("bugfix", ctypes.c_uint8),
        ("build", ctypes.c_uint8),
    ]


class BOA_UuidVersion(ctypes.Structure):
    _fields_ = [("uuid", BOA_UuidBin), ("version", BOA_Version)]


class BOA_ServiceId(ctypes.Structure):
    _fields_ = [("api", BOA_UuidVersion), ("access", BOA_UuidVersion)]


class BOA_ServiceIdParam(ctypes.Structure):
    _fields_ = [
        ("id", BOA_ServiceId),
        ("count", ctypes.c_uint32),
        ("accessParam", ctypes.c_char * 128),
    ]


# Connection Service Interface (CSI)

# CSI - Search For Service (SFS)

CSI_NodeType = ctypes.c_uint32
CSI_NODE_MIN = CSI_NodeType(0)
CSI_NODE_MAX = CSI_NodeType(0x7FFF)


class CSI_NodeRange(ctypes.Structure):
    _fields_ = [("min", CSI_NodeType), ("max", CSI_NodeType)]


class CSI_SubItem(ctypes.Structure):
    _fields_ = [
        ("server", BOA_ServiceIdParam),
        ("nodeType", CSI_NodeType),
        ("uriName", ctypes.c_char * 128),
        ("visibleName", ctypes.c_char * 4),
        ("version", BOA_Version),
        ("reserved2", ctypes.c_char * 88),
        ("serverAffinity", BOA_UuidBin),
        ("requiredAffinity0", BOA_UuidBin),
        ("reserved", ctypes.c_int32 * 4),
        ("requiredAffinity1", BOA_UuidBin),
        ("count", ctypes.c_int32),
        ("requiredAPI", BOA_ServiceId * 4),
    ]


class CSI_Tree(ctypes.Structure):
    pass


CSI_Tree._fields_ = [
    ("item", CSI_SubItem),
    ("sibling", ctypes.POINTER(CSI_Tree)),
    ("child", ctypes.POINTER(CSI_Tree)),
    ("childrenProbed", ctypes.c_int),
]

CSI_CreateProtocolTree = _csi.CSI_CreateProtocolTree
CSI_CreateProtocolTree.argtypes = [
    ctypes.c_char_p,
    CSI_NodeRange,
    ctypes.POINTER(ctypes.POINTER(CSI_Tree)),
]
CSI_CreateProtocolTree.restype = BOA_ResultCode
CSI_CreateProtocolTree.errcheck = errcheck_init

CSI_DestroyProtocolTree = _csi.CSI_DestroyProtocolTree
CSI_DestroyProtocolTree.argtypes = [
    ctypes.POINTER(CSI_Tree),
]
CSI_DestroyProtocolTree.restype = BOA_ResultCode
CSI_DestroyProtocolTree.errcheck = errcheck_oper

# Open Controller Interface (OCI)

# OCI Common - Global Types

OCI_NO_VALUE = BOA_NO_VALUE
OCI_NO_HANDLE = BOA_NO_HANDLE
OCI_Handle = BOA_Handle
OCI_Time = BOA_Time

# OCI Common - Controller Handling

OCI_ControllerHandle = OCI_Handle

OCI_ControllerPropertiesMode = ctypes.c_uint32
OCI_CONTROLLER_MODE_RUNNING = OCI_ControllerPropertiesMode(0)
OCI_CONTROLLER_MODE_SUSPENDED = OCI_ControllerPropertiesMode(1)

OCI_SelfReceptionMode = ctypes.c_uint32
OCI_SELF_RECEPTION_OFF = OCI_SelfReceptionMode(0)
OCI_SELF_RECEPTION_ON = OCI_SelfReceptionMode(1)

# OCI Common - Event Handling

# OCI Common - Error Management

OCI_ErrorCode = BOA_ResultCode

OCI_InternalErrorEvent = ctypes.c_uint32
OCI_INTERNAL_GENERAL_ERROR = OCI_InternalErrorEvent(0)


class OCI_InternalErrorEventMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("eventCode", OCI_InternalErrorEvent),
        ("errorCode", OCI_ErrorCode),
    ]


OCI_GetError = _oci.OCI_GetError
OCI_GetError.argtypes = [
    OCI_Handle,
    OCI_ErrorCode,
    ctypes.c_char_p,
    ctypes.c_uint32,
]
OCI_GetError.restype = OCI_ErrorCode
OCI_GetError.errcheck = errcheck_oper

# OCI Common - Queue Handling

OCI_QueueHandle = OCI_Handle

OCI_QueueEvent = ctypes.c_uint32
OCI_QUEUE_UNDERRUN = OCI_QueueEvent(0)
OCI_QUEUE_EMPTY = OCI_QueueEvent(1)
OCI_QUEUE_NOT_EMPTY = OCI_QueueEvent(2)
OCI_QUEUE_LOW_WATERMARK = OCI_QueueEvent(3)
OCI_QUEUE_HIGH_WATERMARK = OCI_QueueEvent(4)
OCI_QUEUE_FULL = OCI_QueueEvent(5)
OCI_QUEUE_OVERFLOW = OCI_QueueEvent(6)


class OCI_QueueEventMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("eventCode", OCI_QueueEvent),
        ("destination", ctypes.c_uint32),
    ]


OCI_ResetQueue = _oci.OCI_ResetQueue
OCI_ResetQueue.argtypes = [OCI_QueueHandle]
OCI_ResetQueue.restype = OCI_ErrorCode
OCI_ResetQueue.errcheck = errcheck_oper

# OCI Common - Timer Handling

OCI_NO_TIME = BOA_NO_TIME

OCI_TimeReferenceScale = ctypes.c_uint32
OCI_TimeReferenceScaleUnknown = OCI_TimeReferenceScale(0)
OCI_TimeReferenceScaleTAI = OCI_TimeReferenceScale(1)
OCI_TimeReferenceScaleUTC = OCI_TimeReferenceScale(2)

OCI_TimerEvent = ctypes.c_uint32
OCI_TIMER_EVENT_SYNC_LOCK = OCI_TimerEvent(0)
OCI_TIMER_EVENT_SYNC_LOSS = OCI_TimerEvent(1)
OCI_TIMER_EVENT_LEAP_SECOND = OCI_TimerEvent(2)


class OCI_TimerCapabilities(ctypes.Structure):
    _fields_ = [
        ("localClockID", ctypes.c_char * 40),
        ("format", ctypes.c_uint32),
        ("tickFrequency", ctypes.c_uint32),
        ("ticksPerIncrement", ctypes.c_uint32),
        ("localStratumLevel", ctypes.c_uint32),
        ("localReferenceScale", OCI_TimeReferenceScale),
        ("localTimeOriginIso8601", ctypes.c_char * 40),
        ("syncSlave", ctypes.c_uint32),
        ("syncMaster", ctypes.c_uint32),
    ]


class OCI_TimerEventMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("eventCode", OCI_TimerEvent),
        ("destination", ctypes.c_uint32),
    ]


OCI_GetTimerCapabilities = _oci.OCI_GetTimerCapabilities
OCI_GetTimerCapabilities.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_TimerCapabilities),
]
OCI_GetTimerCapabilities.restype = OCI_ErrorCode
OCI_GetTimerCapabilities.errcheck = errcheck_init

OCI_GetTimerValue = _oci.OCI_GetTimerValue
OCI_GetTimerValue.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_Time),
]
OCI_GetTimerValue.restype = OCI_ErrorCode
OCI_GetTimerValue.errcheck = errcheck_oper

# OCI CAN

# OCI CAN - CAN-FD

OCI_CANFDRxMode = ctypes.c_uint32
OCI_CANFDRXMODE_CANFD_FRAMES_IGNORED = OCI_CANFDRxMode(1)
OCI_CANFDRXMODE_CANFD_FRAMES_USING_CAN_MESSAGE = OCI_CANFDRxMode(2)
OCI_CANFDRXMODE_CANFD_FRAMES_USING_CANFD_MESSAGE = OCI_CANFDRxMode(4)
OCI_CANFDRXMODE_CANFD_FRAMES_USING_CANFD_MESSAGE_PADDING = OCI_CANFDRxMode(8)

OCI_CANRxMode = ctypes.c_uint32
OCI_CAN_RXMODE_CAN_FRAMES_IGNORED = OCI_CANRxMode(1)
OCI_CAN_RXMODE_CAN_FRAMES_USING_CAN_MESSAGE = OCI_CANRxMode(2)

OCI_CANFDTxConfig = ctypes.c_uint32
OCI_CANFDTX_USE_CAN_FRAMES_ONLY = OCI_CANFDTxConfig(1)
OCI_CANFDTX_USE_CANFD_FRAMES_ONLY = OCI_CANFDTxConfig(2)
OCI_CANFDTX_USE_CAN_AND_CANFD_FRAMES = OCI_CANFDTxConfig(4)


class OCI_CANFDRxConfig(ctypes.Structure):
    _fields_ = [
        ("canRxMode", OCI_CANRxMode),
        ("canFdRxMode", OCI_CANFDRxMode),
    ]


class OCI_CANFDConfiguration(ctypes.Structure):
    _fields_ = [
        ("dataBitRate", ctypes.c_uint32),
        ("dataSamplePoint", ctypes.c_uint32),
        ("dataBTL_Cycles", ctypes.c_uint32),
        ("dataSJW", ctypes.c_uint32),
        ("flags", ctypes.c_uint32),
        ("txSecondarySamplePointOffset", ctypes.c_uint32),
        ("canFdRxConfig", OCI_CANFDRxConfig),
        ("canFdTxConfig", OCI_CANFDTxConfig),
        ("txSecondarySamplePointFilterWindow", ctypes.c_uint16),
        ("reserved", ctypes.c_uint16),
    ]


class OCI_CANFDRxMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("frameID", ctypes.c_uint32),
        ("flags", ctypes.c_uint16),
        ("res", ctypes.c_uint8),
        ("size", ctypes.c_uint8),
        ("res1", ctypes.c_uint8 * 4),
        ("data", ctypes.c_uint8 * 64),
    ]


class OCI_CANFDTxMessage(ctypes.Structure):
    _fields_ = [
        ("frameID", ctypes.c_uint32),
        ("flags", ctypes.c_uint16),
        ("res", ctypes.c_uint8),
        ("size", ctypes.c_uint8),
        ("data", ctypes.c_uint8 * 64),
    ]


# OCI CAN - Initialization

OCI_CAN_THREE_SAMPLES_PER_BIT = 2
OCI_CAN_SINGLE_SYNC_EDGE = 1
OCI_CAN_MEDIA_HIGH_SPEED = 1

OCI_CAN_STATE_ACTIVE = 0x00000001
OCI_CAN_STATE_PASSIVE = 0x00000002
OCI_CAN_STATE_ERRLIMIT = 0x00000004
OCI_CAN_STATE_BUSOFF = 0x00000008

OCI_CANBusParticipationMode = ctypes.c_uint32
OCI_BUSMODE_PASSIVE = OCI_CANBusParticipationMode(1)
OCI_BUSMODE_ACTIVE = OCI_CANBusParticipationMode(2)

OCI_CANBusTransmissionPolicies = ctypes.c_uint32
OCI_CANTX_UNDEFINED = OCI_CANBusTransmissionPolicies(0)
OCI_CANTX_DONTCARE = OCI_CANBusTransmissionPolicies(0)
OCI_CANTX_FIFO = OCI_CANBusTransmissionPolicies(1)
OCI_CANTX_BESTEFFORT = OCI_CANBusTransmissionPolicies(2)


class OCI_CANConfiguration(ctypes.Structure):
    _fields_ = [
        ("baudrate", ctypes.c_uint32),
        ("samplePoint", ctypes.c_uint32),
        ("samplesPerBit", ctypes.c_uint32),
        ("BTL_Cycles", ctypes.c_uint32),
        ("SJW", ctypes.c_uint32),
        ("syncEdge", ctypes.c_uint32),
        ("physicalMedia", ctypes.c_uint32),
        ("selfReceptionMode", OCI_SelfReceptionMode),
        ("busParticipationMode", OCI_CANBusParticipationMode),
        ("canFDEnabled", ctypes.c_uint32),
        ("canFDConfig", OCI_CANFDConfiguration),
        ("canTxPolicy", OCI_CANBusTransmissionPolicies),
    ]


class OCI_CANControllerProperties(ctypes.Structure):
    _fields_ = [
        ("mode", OCI_ControllerPropertiesMode),
    ]


class OCI_CANControllerCapabilities(ctypes.Structure):
    _fields_ = [
        ("samplesPerBit", ctypes.c_uint32),
        ("syncEdge", ctypes.c_uint32),
        ("physicalMedia", ctypes.c_uint32),
        ("reserved", ctypes.c_uint32),
        ("busEvents", ctypes.c_uint32),
        ("errorFrames", ctypes.c_uint32),
        ("messageFlags", ctypes.c_uint32),
        ("canFDSupport", ctypes.c_uint32),
        ("canFDMaxDataSize", ctypes.c_uint32),
        ("canFDMaxQualifiedDataRate", ctypes.c_uint32),
        ("canFDMaxDataRate", ctypes.c_uint32),
        ("canFDRxConfig_CANMode", ctypes.c_uint32),
        ("canFDRxConfig_CANFDMode", ctypes.c_uint32),
        ("canFDTxConfig_Mode", ctypes.c_uint32),
        ("canBusParticipationMode", ctypes.c_uint32),
        ("canTxPolicies", ctypes.c_uint32),
    ]


class OCI_CANControllerStatus(ctypes.Structure):
    _fields_ = [
        ("reserved", ctypes.c_uint32),
        ("stateCode", ctypes.c_uint32),
    ]


OCI_CreateCANControllerNoSearch = _oci.OCI_CreateCANControllerNoSearch
OCI_CreateCANControllerNoSearch.argtypes = [
    ctypes.c_char_p,
    ctypes.POINTER(BOA_Version),
    ctypes.POINTER(CSI_Tree),
    ctypes.POINTER(OCI_ControllerHandle),
]
OCI_CreateCANControllerNoSearch.restype = OCI_ErrorCode
OCI_CreateCANControllerNoSearch.errcheck = errcheck_init

OCI_OpenCANController = _oci.OCI_OpenCANController
OCI_OpenCANController.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_CANConfiguration),
    ctypes.POINTER(OCI_CANControllerProperties),
]
OCI_OpenCANController.restype = OCI_ErrorCode
# no .errcheck, since we tolerate OCI_WARN_PARAM_ADAPTED warning
# OCI_OpenCANController.errcheck = errcheck_init

OCI_CloseCANController = _oci.OCI_CloseCANController
OCI_CloseCANController.argtypes = [OCI_ControllerHandle]
OCI_CloseCANController.restype = OCI_ErrorCode
OCI_CloseCANController.errcheck = errcheck_oper

OCI_DestroyCANController = _oci.OCI_DestroyCANController
OCI_DestroyCANController.argtypes = [OCI_ControllerHandle]
OCI_DestroyCANController.restype = OCI_ErrorCode
OCI_DestroyCANController.errcheck = errcheck_oper

OCI_AdaptCANConfiguration = _oci.OCI_AdaptCANConfiguration
OCI_AdaptCANConfiguration.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_CANConfiguration),
]
OCI_AdaptCANConfiguration.restype = OCI_ErrorCode
OCI_AdaptCANConfiguration.errcheck = errcheck_oper

OCI_GetCANControllerCapabilities = _oci.OCI_GetCANControllerCapabilities
OCI_GetCANControllerCapabilities.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_CANControllerCapabilities),
]
OCI_GetCANControllerCapabilities.restype = OCI_ErrorCode
OCI_GetCANControllerCapabilities.errcheck = errcheck_init

OCI_GetCANControllerStatus = _oci.OCI_GetCANControllerStatus
OCI_GetCANControllerStatus.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_CANControllerStatus),
]
OCI_GetCANControllerStatus.restype = OCI_ErrorCode
OCI_GetCANControllerStatus.errcheck = errcheck_oper


# OCI CAN - Filter


class OCI_CANRxFilter(ctypes.Structure):
    _fields_ = [
        ("frameIDValue", ctypes.c_uint32),
        ("frameIDMask", ctypes.c_uint32),
        ("tag", ctypes.c_uint32),
    ]


class OCI_CANRxFilterEx(ctypes.Structure):
    _fields_ = [
        ("frameIDValue", ctypes.c_uint32),
        ("frameIDMask", ctypes.c_uint32),
        ("tag", ctypes.c_uint32),
        ("flagsValue", ctypes.c_uint16),
        ("flagsMask", ctypes.c_uint16),
    ]


OCI_AddCANFrameFilterEx = _oci.OCI_AddCANFrameFilterEx
OCI_AddCANFrameFilterEx.argtypes = [
    OCI_QueueHandle,
    ctypes.POINTER(ctypes.POINTER(OCI_CANRxFilterEx)),
    ctypes.c_uint32,
]
OCI_AddCANFrameFilterEx.restype = OCI_ErrorCode
OCI_AddCANFrameFilterEx.errcheck = errcheck_oper

OCI_RemoveCANFrameFilterEx = _oci.OCI_RemoveCANFrameFilterEx
OCI_RemoveCANFrameFilterEx.argtypes = [
    OCI_QueueHandle,
    ctypes.POINTER(ctypes.POINTER(OCI_CANRxFilterEx)),
    ctypes.c_uint32,
]
OCI_RemoveCANFrameFilterEx.restype = OCI_ErrorCode
OCI_RemoveCANFrameFilterEx.errcheck = errcheck_oper

# OCI CAN - Messages

OCI_CAN_MSG_FLAG_EXTENDED = 0x1
OCI_CAN_MSG_FLAG_REMOTE_FRAME = 0x2
OCI_CAN_MSG_FLAG_SELFRECEPTION = 0x4
OCI_CAN_MSG_FLAG_FD_DATA_BIT_RATE = 0x8
OCI_CAN_MSG_FLAG_FD_TRUNC_AND_PAD = 0x10
OCI_CAN_MSG_FLAG_FD_ERROR_PASSIVE = 0x20
OCI_CAN_MSG_FLAG_FD_DATA = 0x40

OCI_CANMessageDataType = ctypes.c_uint32
OCI_CAN_RX_MESSAGE = OCI_CANMessageDataType(1)
OCI_CAN_TX_MESSAGE = OCI_CANMessageDataType(2)
OCI_CAN_ERROR_FRAME = OCI_CANMessageDataType(3)
OCI_CAN_BUS_EVENT = OCI_CANMessageDataType(4)
OCI_CAN_INTERNAL_ERROR_EVENT = OCI_CANMessageDataType(5)
OCI_CAN_QUEUE_EVENT = OCI_CANMessageDataType(6)
OCI_CAN_TIMER_EVENT = OCI_CANMessageDataType(7)
OCI_CANFDRX_MESSAGE = OCI_CANMessageDataType(8)
OCI_CANFDTX_MESSAGE = OCI_CANMessageDataType(9)


class OCI_CANTxMessage(ctypes.Structure):
    _fields_ = [
        ("frameID", ctypes.c_uint32),
        ("flags", ctypes.c_uint16),
        ("res", ctypes.c_uint8),
        ("dlc", ctypes.c_uint8),
        ("data", ctypes.c_uint8 * 8),
    ]


class OCI_CANRxMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("frameID", ctypes.c_uint32),
        ("flags", ctypes.c_uint16),
        ("res", ctypes.c_uint8),
        ("dlc", ctypes.c_uint8),
        ("res1", ctypes.c_uint8 * 4),
        ("data", ctypes.c_uint8 * 8),
    ]


class OCI_CANErrorFrameMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("frameID", ctypes.c_uint32),
        ("flags", ctypes.c_uint16),
        ("res", ctypes.c_uint8),
        ("dlc", ctypes.c_uint8),
        ("type", ctypes.c_uint32),
        ("destination", ctypes.c_uint32),
    ]


class OCI_CANEventMessage(ctypes.Structure):
    _fields_ = [
        ("timeStamp", OCI_Time),
        ("tag", ctypes.c_uint32),
        ("eventCode", ctypes.c_uint32),
        ("destination", ctypes.c_uint32),
    ]


class OCI_CANMessageData(ctypes.Union):
    _fields_ = [
        ("rxMessage", OCI_CANRxMessage),
        ("txMessage", OCI_CANTxMessage),
        ("errorFrameMessage", OCI_CANErrorFrameMessage),
        ("canEventMessage", OCI_CANEventMessage),
        ("internalErrorEventMessage", OCI_InternalErrorEventMessage),
        ("timerEventMessage", OCI_TimerEventMessage),
        ("queueEventMessage", OCI_QueueEventMessage),
    ]


class OCI_CANMessageDataEx(ctypes.Union):
    _fields_ = [
        ("rxMessage", OCI_CANRxMessage),
        ("txMessage", OCI_CANTxMessage),
        ("errorFrameMessage", OCI_CANErrorFrameMessage),
        ("canEventMessage", OCI_CANEventMessage),
        ("internalErrorEventMessage", OCI_InternalErrorEventMessage),
        ("timerEventMessage", OCI_TimerEventMessage),
        ("queueEventMessage", OCI_QueueEventMessage),
        ("canFDRxMessage", OCI_CANFDRxMessage),
        ("canFDTxMessage", OCI_CANFDTxMessage),
    ]


class OCI_CANMessage(ctypes.Structure):
    _fields_ = [
        ("type", OCI_CANMessageDataType),
        ("reserved", ctypes.c_uint32),
        ("data", OCI_CANMessageData),
    ]


class OCI_CANMessageEx(ctypes.Structure):
    _fields_ = [
        ("type", OCI_CANMessageDataType),
        ("reserved", ctypes.c_uint32),
        ("data", OCI_CANMessageDataEx),
    ]


# OCI CAN - Queues

OCI_CANRxCallbackFunctionSingleMsg = ctypes.CFUNCTYPE(
    None, ctypes.c_void_p, ctypes.POINTER(OCI_CANMessage)
)

OCI_CANRxCallbackFunctionSingleMsgEx = ctypes.CFUNCTYPE(
    None, ctypes.c_void_p, ctypes.POINTER(OCI_CANMessageEx)
)


class OCI_CANRxCallbackSingleMsg(ctypes.Structure):
    class _U(ctypes.Union):
        _fields_ = [
            ("function", OCI_CANRxCallbackFunctionSingleMsg),
            ("functionEx", OCI_CANRxCallbackFunctionSingleMsgEx),
        ]

    _anonymous_ = ("u",)
    _fields_ = [
        ("u", _U),
        ("userData", ctypes.c_void_p),
    ]


class OCI_CANRxQueueConfiguration(ctypes.Structure):
    _fields_ = [
        ("onFrame", OCI_CANRxCallbackSingleMsg),
        ("onEvent", OCI_CANRxCallbackSingleMsg),
        ("selfReceptionMode", OCI_SelfReceptionMode),
    ]


class OCI_CANTxQueueConfiguration(ctypes.Structure):
    _fields_ = [
        ("reserved", ctypes.c_uint32),
    ]


OCI_CreateCANRxQueue = _oci.OCI_CreateCANRxQueue
OCI_CreateCANRxQueue.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_CANRxQueueConfiguration),
    ctypes.POINTER(OCI_QueueHandle),
]
OCI_CreateCANRxQueue.restype = OCI_ErrorCode
OCI_CreateCANRxQueue.errcheck = errcheck_init

OCI_DestroyCANRxQueue = _oci.OCI_DestroyCANRxQueue
OCI_DestroyCANRxQueue.argtypes = [OCI_QueueHandle]
OCI_DestroyCANRxQueue.restype = OCI_ErrorCode
OCI_DestroyCANRxQueue.errcheck = errcheck_oper

OCI_CreateCANTxQueue = _oci.OCI_CreateCANTxQueue
OCI_CreateCANTxQueue.argtypes = [
    OCI_ControllerHandle,
    ctypes.POINTER(OCI_CANTxQueueConfiguration),
    ctypes.POINTER(OCI_QueueHandle),
]
OCI_CreateCANTxQueue.restype = OCI_ErrorCode
OCI_CreateCANTxQueue.errcheck = errcheck_init

OCI_DestroyCANTxQueue = _oci.OCI_DestroyCANTxQueue
OCI_DestroyCANTxQueue.argtypes = [OCI_QueueHandle]
OCI_DestroyCANTxQueue.restype = OCI_ErrorCode
OCI_DestroyCANTxQueue.errcheck = errcheck_oper

OCI_WriteCANDataEx = _oci.OCI_WriteCANDataEx
OCI_WriteCANDataEx.argtypes = [
    OCI_QueueHandle,
    OCI_Time,
    ctypes.POINTER(ctypes.POINTER(OCI_CANMessageEx)),
    ctypes.c_uint32,
    ctypes.POINTER(ctypes.c_uint32),
]
OCI_WriteCANDataEx.restype = OCI_ErrorCode
OCI_WriteCANDataEx.errcheck = errcheck_oper

OCI_ReadCANDataEx = _oci.OCI_ReadCANDataEx
OCI_ReadCANDataEx.argtypes = [
    OCI_QueueHandle,
    OCI_Time,
    ctypes.POINTER(ctypes.POINTER(OCI_CANMessageEx)),
    ctypes.c_uint32,
    ctypes.POINTER(ctypes.c_uint32),
    ctypes.POINTER(ctypes.c_uint32),
]
OCI_ReadCANDataEx.restype = OCI_ErrorCode
OCI_ReadCANDataEx.errcheck = errcheck_oper



import logging
from threading import Event

from can import BusABC, BusState, Message

from ...exceptions import CanError, CanInitializationError, CanOperationError
from .constants import *
from .exceptions import UcanException
from .structures import *
from .ucan import UcanServer

log = logging.getLogger("can.systec")


class Ucan(UcanServer):
    """
    Wrapper around UcanServer to read messages with timeout using events.
    """

    def __init__(self):
        super().__init__()
        self._msg_received_event = Event()

    def can_msg_received_event(self, channel):
        self._msg_received_event.set()

    def read_can_msg(self, channel, count, timeout):
        self._msg_received_event.clear()
        if self.get_msg_pending(channel, PendingFlags.PENDING_FLAG_RX_DLL) == 0:
            if not self._msg_received_event.wait(timeout):
                return None, False
        return super().read_can_msg(channel, 1)


class UcanBus(BusABC):
    """
    The CAN Bus implemented for the SYSTEC interface.
    """

    BITRATES = {
        10000: Baudrate.BAUD_10kBit,
        20000: Baudrate.BAUD_20kBit,
        50000: Baudrate.BAUD_50kBit,
        100000: Baudrate.BAUD_100kBit,
        125000: Baudrate.BAUD_125kBit,
        250000: Baudrate.BAUD_250kBit,
        500000: Baudrate.BAUD_500kBit,
        800000: Baudrate.BAUD_800kBit,
        1000000: Baudrate.BAUD_1MBit,
    }

    def __init__(self, channel, can_filters=None, **kwargs):
        """
        :param int channel:
            The Channel id to create this bus with.

        :param list can_filters:
            See :meth:`can.BusABC.set_filters`.

        Backend Configuration

        :param int bitrate:
            Channel bitrate in bit/s.
            Default is 500000.

        :param int device_number:
            The device number of the USB-CAN.
            Valid values: 0 through 254. Special value 255 is reserved to detect the first connected device (should only
            be used, in case only one module is connected to the computer).
            Default is 255.

        :param can.bus.BusState state:
            BusState of the channel.
            Default is ACTIVE.

        :param bool receive_own_messages:
            If messages transmitted should also be received back.
            Default is False.

        :param int rx_buffer_entries:
            The maximum number of entries in the receive buffer.
            Default is 4096.

        :param int tx_buffer_entries:
            The maximum number of entries in the transmit buffer.
            Default is 4096.

        :raises ValueError:
            If invalid input parameter were passed.

        :raises ~can.exceptions.CanInterfaceNotImplementedError:
            If the platform is not supported.

        :raises ~can.exceptions.CanInitializationError:
            If hardware or CAN interface initialization failed.
        """
        try:
            self._ucan = Ucan()
        except CanError as error:
            raise error
        except Exception as exception:
            raise CanInitializationError(
                "The SYSTEC ucan library has not been initialized."
            ) from exception

        self.channel = int(channel)
        device_number = int(kwargs.get("device_number", ANY_MODULE))

        # configuration options
        bitrate = kwargs.get("bitrate", 500000)
        if bitrate not in self.BITRATES:
            raise ValueError(f"Invalid bitrate {bitrate}")

        state = kwargs.get("state", BusState.ACTIVE)
        if state is BusState.ACTIVE or state is BusState.PASSIVE:
            self._state = state
        else:
            raise ValueError("BusState must be Active or Passive")

        # get parameters
        self._params = {
            "mode": Mode.MODE_NORMAL
            | (Mode.MODE_TX_ECHO if kwargs.get("receive_own_messages") else 0)
            | (Mode.MODE_LISTEN_ONLY if state is BusState.PASSIVE else 0),
            "BTR": self.BITRATES[bitrate],
        }
        # get extra parameters
        if kwargs.get("rx_buffer_entries"):
            self._params["rx_buffer_entries"] = int(kwargs.get("rx_buffer_entries"))
        if kwargs.get("tx_buffer_entries"):
            self._params["tx_buffer_entries"] = int(kwargs.get("tx_buffer_entries"))

        try:
            self._ucan.init_hardware(device_number=device_number)
            self._ucan.init_can(self.channel, **self._params)
            hw_info_ex, _, _ = self._ucan.get_hardware_info()
            self.channel_info = "{}, S/N {}, CH {}, BTR {}".format(
                self._ucan.get_product_code_message(hw_info_ex.product_code),
                hw_info_ex.serial,
                self.channel,
                self._ucan.get_baudrate_message(self.BITRATES[bitrate]),
            )
        except UcanException as exception:
            raise CanInitializationError() from exception

        self._is_filtered = False

        super().__init__(channel=channel, can_filters=can_filters, **kwargs)

    def _recv_internal(self, timeout):
        try:
            message, _ = self._ucan.read_can_msg(self.channel, 1, timeout)
        except UcanException as exception:
            raise CanOperationError() from exception

        if not message:
            return None, False

        msg = Message(
            timestamp=float(message[0].time) / 1000.0,
            is_remote_frame=bool(message[0].frame_format & MsgFrameFormat.MSG_FF_RTR),
            is_extended_id=bool(message[0].frame_format & MsgFrameFormat.MSG_FF_EXT),
            arbitration_id=message[0].id,
            dlc=len(message[0].data),
            data=message[0].data,
        )
        return msg, self._is_filtered

    def send(self, msg, timeout=None):
        """
        Sends one CAN message.

        When a transmission timeout is set the firmware tries to send
        a message within this timeout. If it could not be sent the firmware sets
        the "auto delete" state. Within this state all transmit CAN messages for
        this channel will be deleted automatically for not blocking the other channel.

        :param can.Message msg:
            The CAN message.

        :param float timeout:
            Transmit timeout in seconds (value 0 switches off the "auto delete")

        :raises ~can.exceptions.CanOperationError:
            If the message could not be sent.
        """
        try:
            if timeout is not None and timeout >= 0:
                self._ucan.set_tx_timeout(self.channel, int(timeout * 1000))

            message = CanMsg(
                msg.arbitration_id,
                MsgFrameFormat.MSG_FF_STD
                | (MsgFrameFormat.MSG_FF_EXT if msg.is_extended_id else 0)
                | (MsgFrameFormat.MSG_FF_RTR if msg.is_remote_frame else 0),
                msg.data,
            )
            self._ucan.write_can_msg(self.channel, [message])
        except UcanException as exception:
            raise CanOperationError() from exception

    @staticmethod
    def _detect_available_configs():
        configs = []
        try:
            # index, is_used, hw_info_ex, init_info
            for _, _, hw_info_ex, _ in Ucan.enumerate_hardware():
                configs.append(
                    {
                        "interface": "systec",
                        "channel": Channel.CHANNEL_CH0,
                        "device_number": hw_info_ex.device_number,
                    }
                )
                if Ucan.check_support_two_channel(hw_info_ex):
                    configs.append(
                        {
                            "interface": "systec",
                            "channel": Channel.CHANNEL_CH1,
                            "device_number": hw_info_ex.device_number,
                        }
                    )
        except Exception:
            log.warning("The SYSTEC ucan library has not been initialized.")
        return configs

    def _apply_filters(self, filters):
        try:
            if filters and len(filters) == 1:
                can_id = filters[0]["can_id"]
                can_mask = filters[0]["can_mask"]
                self._ucan.set_acceptance(self.channel, can_mask, can_id)
                self._is_filtered = True
                log.info("Hardware filtering on ID 0x%X, mask 0x%X", can_id, can_mask)
            else:
                self._ucan.set_acceptance(self.channel)
                self._is_filtered = False
                log.info("Hardware filtering has been disabled")
        except UcanException as exception:
            raise CanOperationError() from exception

    def flush_tx_buffer(self):
        """
        Flushes the transmit buffer.

        :raises ~can.exceptions.CanError:
            If flushing of the transmit buffer failed.
        """
        log.info("Flushing transmit buffer")
        try:
            self._ucan.reset_can(self.channel, ResetFlags.RESET_ONLY_TX_BUFF)
        except UcanException as exception:
            raise CanOperationError() from exception

    @staticmethod
    def create_filter(extended, from_id, to_id, rtr_only, rtr_too):
        """
        Calculates AMR and ACR using CAN-ID as parameter.

        :param bool extended:
            if True parameters from_id and to_id contains 29-bit CAN-ID

        :param int from_id:
            first CAN-ID which should be received

        :param int to_id:
            last CAN-ID which should be received

        :param bool rtr_only:
            if True only RTR-Messages should be received, and rtr_too will be ignored

        :param bool rtr_too:
            if True CAN data frames and RTR-Messages should be received

        :return: Returns list with one filter containing a "can_id", a "can_mask" and "extended" key.
        """
        return [
            {
                "can_id": Ucan.calculate_acr(
                    extended, from_id, to_id, rtr_only, rtr_too
                ),
                "can_mask": Ucan.calculate_amr(
                    extended, from_id, to_id, rtr_only, rtr_too
                ),
                "extended": extended,
            }
        ]

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        if self._state is not BusState.ERROR and (
            new_state is BusState.ACTIVE or new_state is BusState.PASSIVE
        ):
            try:
                # close the CAN channel
                self._ucan.shutdown(self.channel, False)
                # set mode
                if new_state is BusState.ACTIVE:
                    self._params["mode"] &= ~Mode.MODE_LISTEN_ONLY
                else:
                    self._params["mode"] |= Mode.MODE_LISTEN_ONLY
                # reinitialize CAN channel
                self._ucan.init_can(self.channel, **self._params)
            except UcanException as exception:
                raise CanOperationError() from exception

    def shutdown(self):
        """
        Shuts down all CAN interfaces and hardware interface.
        """
        super().shutdown()
        try:
            self._ucan.shutdown()
        except Exception as exception:
            log.error(exception)



"""
Ctypes wrapper module for IXXAT Virtual CAN Interface V4 on win32 systems

Copyright (C) 2016 Giuseppe Corbelli <giuseppe.corbelli@weightpack.com>
Copyright (C) 2019 Marcel Kanter <marcel.kanter@googlemail.com>
"""

from can import (
    CanInitializationError,
    CanOperationError,
    CanTimeoutError,
)

__all__ = [
    "VCITimeout",
    "VCIError",
    "VCIRxQueueEmptyError",
    "VCIBusOffError",
    "VCIDeviceNotFoundError",
]


class VCITimeout(CanTimeoutError):
    """Wraps the VCI_E_TIMEOUT error"""


class VCIError(CanOperationError):
    """Try to display errors that occur within the wrapped C library nicely."""


class VCIRxQueueEmptyError(VCIError):
    """Wraps the VCI_E_RXQUEUE_EMPTY error"""

    def __init__(self):
        super().__init__("Receive queue is empty")


class VCIBusOffError(VCIError):
    def __init__(self):
        super().__init__("Controller is in BUSOFF state")


class VCIDeviceNotFoundError(CanInitializationError):
    pass



#!/usr/bin/env python

"""
Test rotating loggers
"""

import os
from pathlib import Path
from unittest.mock import Mock

import can

from .data.example_data import generate_message


class TestBaseRotatingLogger:
    @staticmethod
    def _get_instance(path, *args, **kwargs) -> can.io.BaseRotatingLogger:
        class SubClass(can.io.BaseRotatingLogger):
            """Subclass that implements abstract methods for testing."""

            _supported_formats = {".asc", ".blf", ".csv", ".log", ".txt"}

            def __init__(self, *args, **kwargs) -> None:
                super().__init__(*args, **kwargs)
                self._writer = can.Printer(file=path / "__unused.txt")

            def should_rollover(self, msg: can.Message) -> bool:
                return False

            def do_rollover(self):
                ...

        return SubClass(*args, **kwargs)

    def test_import(self):
        assert hasattr(can.io, "BaseRotatingLogger")

    def test_attributes(self):
        assert issubclass(can.io.BaseRotatingLogger, can.Listener)
        assert hasattr(can.io.BaseRotatingLogger, "namer")
        assert hasattr(can.io.BaseRotatingLogger, "rotator")
        assert hasattr(can.io.BaseRotatingLogger, "rollover_count")
        assert hasattr(can.io.BaseRotatingLogger, "writer")
        assert hasattr(can.io.BaseRotatingLogger, "rotation_filename")
        assert hasattr(can.io.BaseRotatingLogger, "rotate")
        assert hasattr(can.io.BaseRotatingLogger, "on_message_received")
        assert hasattr(can.io.BaseRotatingLogger, "stop")
        assert hasattr(can.io.BaseRotatingLogger, "should_rollover")
        assert hasattr(can.io.BaseRotatingLogger, "do_rollover")

    def test_get_new_writer(self, tmp_path):
        logger_instance = self._get_instance(tmp_path)

        writer = logger_instance._get_new_writer(tmp_path / "file.ASC")
        assert isinstance(writer, can.ASCWriter)
        writer.stop()

        writer = logger_instance._get_new_writer(tmp_path / "file.BLF")
        assert isinstance(writer, can.BLFWriter)
        writer.stop()

        writer = logger_instance._get_new_writer(tmp_path / "file.CSV")
        assert isinstance(writer, can.CSVWriter)
        writer.stop()

        writer = logger_instance._get_new_writer(tmp_path / "file.LOG")
        assert isinstance(writer, can.CanutilsLogWriter)
        writer.stop()

        writer = logger_instance._get_new_writer(tmp_path / "file.TXT")
        assert isinstance(writer, can.Printer)
        writer.stop()

    def test_rotation_filename(self, tmp_path):
        logger_instance = self._get_instance(tmp_path)

        default_name = "default"
        assert logger_instance.rotation_filename(default_name) == "default"

        logger_instance.namer = lambda x: x + "_by_namer"
        assert logger_instance.rotation_filename(default_name) == "default_by_namer"

    def test_rotate_without_rotator(self, tmp_path):
        logger_instance = self._get_instance(tmp_path)

        source = str(tmp_path / "source.txt")
        dest = str(tmp_path / "dest.txt")

        assert os.path.exists(source) is False
        assert os.path.exists(dest) is False

        logger_instance._writer = logger_instance._get_new_writer(source)
        logger_instance.stop()

        assert os.path.exists(source) is True
        assert os.path.exists(dest) is False

        logger_instance.rotate(source, dest)

        assert os.path.exists(source) is False
        assert os.path.exists(dest) is True

    def test_rotate_with_rotator(self, tmp_path):
        logger_instance = self._get_instance(tmp_path)

        rotator_func = Mock()
        logger_instance.rotator = rotator_func

        source = str(tmp_path / "source.txt")
        dest = str(tmp_path / "dest.txt")

        assert os.path.exists(source) is False
        assert os.path.exists(dest) is False

        logger_instance._writer = logger_instance._get_new_writer(source)
        logger_instance.stop()

        assert os.path.exists(source) is True
        assert os.path.exists(dest) is False

        logger_instance.rotate(source, dest)
        rotator_func.assert_called_with(source, dest)

        # assert that no rotation was performed since rotator_func
        # does not do anything
        assert os.path.exists(source) is True
        assert os.path.exists(dest) is False

    def test_stop(self, tmp_path):
        """Test if stop() method of writer is called."""
        with self._get_instance(tmp_path) as logger_instance:
            logger_instance._writer = logger_instance._get_new_writer(
                tmp_path / "file.ASC"
            )

            # replace stop method of writer with Mock
            original_stop = logger_instance.writer.stop
            mock_stop = Mock()
            logger_instance.writer.stop = mock_stop

            logger_instance.stop()
            mock_stop.assert_called()

            # close file.ASC to enable cleanup of temp_dir
            original_stop()

    def test_on_message_received(self, tmp_path):
        logger_instance = self._get_instance(tmp_path)

        logger_instance._writer = logger_instance._get_new_writer(tmp_path / "file.ASC")

        # Test without rollover
        should_rollover = Mock(return_value=False)
        do_rollover = Mock()
        writers_on_message_received = Mock()

        logger_instance.should_rollover = should_rollover
        logger_instance.do_rollover = do_rollover
        logger_instance.writer.on_message_received = writers_on_message_received

        msg = generate_message(0x123)
        logger_instance.on_message_received(msg)

        should_rollover.assert_called_with(msg)
        do_rollover.assert_not_called()
        writers_on_message_received.assert_called_with(msg)

        # Test with rollover
        should_rollover = Mock(return_value=True)
        do_rollover = Mock()
        writers_on_message_received = Mock()

        logger_instance.should_rollover = should_rollover
        logger_instance.do_rollover = do_rollover
        logger_instance.writer.on_message_received = writers_on_message_received

        msg = generate_message(0x123)
        logger_instance.on_message_received(msg)

        should_rollover.assert_called_with(msg)
        do_rollover.assert_called()
        writers_on_message_received.assert_called_with(msg)

        # stop writer to enable cleanup of temp_dir
        logger_instance.stop()


class TestSizedRotatingLogger:
    def test_import(self):
        assert hasattr(can.io, "SizedRotatingLogger")
        assert hasattr(can, "SizedRotatingLogger")

    def test_attributes(self):
        assert issubclass(can.SizedRotatingLogger, can.io.BaseRotatingLogger)
        assert hasattr(can.SizedRotatingLogger, "namer")
        assert hasattr(can.SizedRotatingLogger, "rotator")
        assert hasattr(can.SizedRotatingLogger, "should_rollover")
        assert hasattr(can.SizedRotatingLogger, "do_rollover")

    def test_create_instance(self, tmp_path):
        base_filename = "mylogfile.ASC"
        max_bytes = 512

        logger_instance = can.SizedRotatingLogger(
            base_filename=tmp_path / base_filename, max_bytes=max_bytes
        )
        assert Path(logger_instance.base_filename).name == base_filename
        assert logger_instance.max_bytes == max_bytes
        assert logger_instance.rollover_count == 0
        assert isinstance(logger_instance.writer, can.ASCWriter)

        logger_instance.stop()

    def test_should_rollover(self, tmp_path):
        base_filename = "mylogfile.ASC"
        max_bytes = 512

        logger_instance = can.SizedRotatingLogger(
            base_filename=tmp_path / base_filename, max_bytes=max_bytes
        )
        msg = generate_message(0x123)
        do_rollover = Mock()
        logger_instance.do_rollover = do_rollover

        logger_instance.writer.file.tell = Mock(return_value=511)
        assert logger_instance.should_rollover(msg) is False
        logger_instance.on_message_received(msg)
        do_rollover.assert_not_called()

        logger_instance.writer.file.tell = Mock(return_value=512)
        assert logger_instance.should_rollover(msg) is True
        logger_instance.on_message_received(msg)
        do_rollover.assert_called()

        logger_instance.stop()

    def test_logfile_size(self, tmp_path):
        base_filename = "mylogfile.ASC"
        max_bytes = 1024
        msg = generate_message(0x123)

        logger_instance = can.SizedRotatingLogger(
            base_filename=tmp_path / base_filename, max_bytes=max_bytes
        )
        for _ in range(128):
            logger_instance.on_message_received(msg)

        for file_path in os.listdir(tmp_path):
            assert os.path.getsize(tmp_path / file_path) <= 1100

        logger_instance.stop()

    def test_logfile_size_context_manager(self, tmp_path):
        base_filename = "mylogfile.ASC"
        max_bytes = 1024
        msg = generate_message(0x123)

        with can.SizedRotatingLogger(
            base_filename=tmp_path / base_filename, max_bytes=max_bytes
        ) as logger_instance:
            for _ in range(128):
                logger_instance.on_message_received(msg)

            for file_path in os.listdir(tmp_path):
                assert os.path.getsize(os.path.join(tmp_path, file_path)) <= 1100



"""
This module contains the base implementation of :class:`can.BusABC` as well
as a list of all available backends and some implemented
CyclicSendTasks.
"""

import importlib
import logging
from typing import Any, Iterable, List, Optional, Type, Union, cast

from . import util
from .bus import BusABC
from .exceptions import CanInterfaceNotImplementedError
from .interfaces import BACKENDS
from .typechecking import AutoDetectedConfig, Channel

log = logging.getLogger("can.interface")
log_autodetect = log.getChild("detect_available_configs")


def _get_class_for_interface(interface: str) -> Type[BusABC]:
    """
    Returns the main bus class for the given interface.

    :raises:
        NotImplementedError if the interface is not known
    :raises CanInterfaceNotImplementedError:
         if there was a problem while importing the interface or the bus class within that
    """
    # Find the correct backend
    try:
        module_name, class_name = BACKENDS[interface]
    except KeyError:
        raise NotImplementedError(
            f"CAN interface '{interface}' not supported"
        ) from None

    # Import the correct interface module
    try:
        module = importlib.import_module(module_name)
    except Exception as e:
        raise CanInterfaceNotImplementedError(
            f"Cannot import module {module_name} for CAN interface '{interface}': {e}"
        ) from None

    # Get the correct class
    try:
        bus_class = getattr(module, class_name)
    except Exception as e:
        raise CanInterfaceNotImplementedError(
            f"Cannot import class {class_name} from module {module_name} for CAN interface "
            f"'{interface}': {e}"
        ) from None

    return cast(Type[BusABC], bus_class)


class Bus(BusABC):  # pylint: disable=abstract-method
    """Bus wrapper with configuration loading.

    Instantiates a CAN Bus of the given ``interface``, falls back to reading a
    configuration file from default locations.

    .. note::
        Please note that while the arguments provided to this class take precedence
        over any existing values from configuration, it is possible that other parameters
        from the configuration may be added to the bus instantiation.
        This could potentially have unintended consequences. To prevent this,
        you may use the *ignore_config* parameter to ignore any existing configurations.

    :param channel:
        Channel identification. Expected type is backend dependent.
        Set to ``None`` to let it be resolved automatically from the default
        :ref:`configuration`.

    :param interface:
        See :ref:`interface names` for a list of supported interfaces.
        Set to ``None`` to let it be resolved automatically from the default
        :ref:`configuration`.

    :param config_context:
        Extra 'context', that is passed to config sources.
        This can be used to select a section other than 'default' in the configuration file.

    :param ignore_config:
        If ``True``, only the given arguments will be used for the bus instantiation. Existing
        configuration sources will be ignored.

    :param kwargs:
        ``interface`` specific keyword arguments.

    :raises ~can.exceptions.CanInterfaceNotImplementedError:
        if the ``interface`` isn't recognized or cannot be loaded

    :raises ~can.exceptions.CanInitializationError:
        if the bus cannot be instantiated

    :raises ValueError:
        if the ``channel`` could not be determined
    """

    @staticmethod
    @util.deprecated_args_alias(
        deprecation_start="4.2.0",
        deprecation_end="5.0.0",
        bustype="interface",
        context="config_context",
    )
    def __new__(  # type: ignore
        cls: Any,
        channel: Optional[Channel] = None,
        interface: Optional[str] = None,
        config_context: Optional[str] = None,
        ignore_config: bool = False,
        **kwargs: Any,
    ) -> BusABC:
        # figure out the rest of the configuration; this might raise an error
        if interface is not None:
            kwargs["interface"] = interface
        if channel is not None:
            kwargs["channel"] = channel

        if not ignore_config:
            kwargs = util.load_config(config=kwargs, context=config_context)

        # resolve the bus class to use for that interface
        cls = _get_class_for_interface(kwargs["interface"])

        # remove the "interface" key, so it doesn't get passed to the backend
        del kwargs["interface"]

        # make sure the bus can handle this config format
        channel = kwargs.pop("channel", channel)
        if channel is None:
            # Use the default channel for the backend
            bus = cls(**kwargs)
        else:
            bus = cls(channel, **kwargs)

        return cast(BusABC, bus)


def detect_available_configs(
    interfaces: Union[None, str, Iterable[str]] = None
) -> List[AutoDetectedConfig]:
    """Detect all configurations/channels that the interfaces could
    currently connect with.

    This might be quite time consuming.

    Automated configuration detection may not be implemented by
    every interface on every platform. This method will not raise
    an error in that case, but with rather return an empty list
    for that interface.

    :param interfaces: either
        - the name of an interface to be searched in as a string,
        - an iterable of interface names to search in, or
        - `None` to search in all known interfaces.
    :rtype: list[dict]
    :return: an iterable of dicts, each suitable for usage in
             the constructor of :class:`can.BusABC`.
    """

    # Figure out where to search
    if interfaces is None:
        interfaces = BACKENDS
    elif isinstance(interfaces, str):
        interfaces = (interfaces,)
    # else it is supposed to be an iterable of strings

    result = []
    for interface in interfaces:
        try:
            bus_class = _get_class_for_interface(interface)
        except CanInterfaceNotImplementedError:
            log_autodetect.debug(
                'interface "%s" cannot be loaded for detection of available configurations',
                interface,
            )
            continue

        # get available channels
        try:
            available = list(
                bus_class._detect_available_configs()  # pylint: disable=protected-access
            )
        except NotImplementedError:
            log_autodetect.debug(
                'interface "%s" does not support detection of available configurations',
                interface,
            )
        else:
            log_autodetect.debug(
                'interface "%s" detected %i available configurations',
                interface,
                len(available),
            )

            # add the interface name to the configs if it is not already present
            for config in available:
                if "interface" not in config:
                    config["interface"] = interface

            # append to result
            result += available

    return result



"""
Ctypes wrapper module for IXXAT Virtual CAN Interface V4 on win32 systems

Copyright (C) 2016-2021 Giuseppe Corbelli <giuseppe.corbelli@weightpack.com>
"""

__all__ = [
    "get_ixxat_hwids",
    "IXXATBus",
]

from can.interfaces.ixxat.canlib import IXXATBus

# import this and not the one from vcinpl2 for backward compatibility
from can.interfaces.ixxat.canlib_vcinpl import get_ixxat_hwids



"""
This module contains the generic :class:`LogReader` as
well as :class:`MessageSync` which plays back messages
in the recorded order and time intervals.
"""
import gzip
import pathlib
import time
import typing

from pkg_resources import iter_entry_points

from ..message import Message
from ..typechecking import AcceptedIOType, FileLike, StringPathLike
from .asc import ASCReader
from .blf import BLFReader
from .canutils import CanutilsLogReader
from .csv import CSVReader
from .generic import MessageReader
from .mf4 import MF4Reader
from .sqlite import SqliteReader
from .trc import TRCReader


class LogReader(MessageReader):
    """
    Replay logged CAN messages from a file.

    The format is determined from the file suffix which can be one of:
      * .asc
      * .blf
      * .csv
      * .db
      * .log
      * .mf4 (optional, depends on asammdf)
      * .trc

    Gzip compressed files can be used as long as the original
    files suffix is one of the above (e.g. filename.asc.gz).


    Exposes a simple iterator interface, to use simply:

        >>> for msg in LogReader("some/path/to/my_file.log"):
        ...     print(msg)

    .. note::
        There are no time delays, if you want to reproduce the measured
        delays between messages look at the :class:`can.MessageSync` class.

    .. note::
        This class itself is just a dispatcher, and any positional an keyword
        arguments are passed on to the returned instance.
    """

    fetched_plugins = False
    message_readers: typing.Dict[str, typing.Optional[typing.Type[MessageReader]]] = {
        ".asc": ASCReader,
        ".blf": BLFReader,
        ".csv": CSVReader,
        ".db": SqliteReader,
        ".log": CanutilsLogReader,
        ".mf4": MF4Reader,
        ".trc": TRCReader,
    }

    @staticmethod
    def __new__(  # type: ignore
        cls: typing.Any,
        filename: StringPathLike,
        **kwargs: typing.Any,
    ) -> MessageReader:
        """
        :param filename: the filename/path of the file to read from
        :raises ValueError: if the filename's suffix is of an unknown file type
        """
        if not LogReader.fetched_plugins:
            LogReader.message_readers.update(
                {
                    reader.name: reader.load()
                    for reader in iter_entry_points("can.io.message_reader")
                }
            )
            LogReader.fetched_plugins = True

        suffix = pathlib.PurePath(filename).suffix.lower()

        file_or_filename: AcceptedIOType = filename
        if suffix == ".gz":
            suffix, file_or_filename = LogReader.decompress(filename)
        try:
            ReaderType = LogReader.message_readers[suffix]
        except KeyError:
            raise ValueError(
                f'No read support for this unknown log format "{suffix}"'
            ) from None
        if ReaderType is None:
            raise ImportError(f"failed to import reader for extension {suffix}")
        return ReaderType(file=file_or_filename, **kwargs)

    @staticmethod
    def decompress(
        filename: StringPathLike,
    ) -> typing.Tuple[str, typing.Union[str, FileLike]]:
        """
        Return the suffix and io object of the decompressed file.
        """
        real_suffix = pathlib.Path(filename).suffixes[-2].lower()
        mode = "rb" if real_suffix == ".blf" else "rt"

        return real_suffix, gzip.open(filename, mode)

    def __iter__(self) -> typing.Generator[Message, None, None]:
        raise NotImplementedError()


class MessageSync:
    """
    Used to iterate over some given messages in the recorded time.
    """

    def __init__(
        self,
        messages: typing.Iterable[Message],
        timestamps: bool = True,
        gap: float = 0.0001,
        skip: float = 60.0,
    ) -> None:
        """Creates an new **MessageSync** instance.

        :param messages: An iterable of :class:`can.Message` instances.
        :param timestamps: Use the messages' timestamps. If False, uses the *gap* parameter
                           as the time between messages.
        :param gap: Minimum time between sent messages in seconds
        :param skip: Skip periods of inactivity greater than this (in seconds).
        """
        self.raw_messages = messages
        self.timestamps = timestamps
        self.gap = gap
        self.skip = skip

    def __iter__(self) -> typing.Generator[Message, None, None]:
        t_wakeup = playback_start_time = time.perf_counter()
        recorded_start_time = None
        t_skipped = 0.0

        for message in self.raw_messages:
            # Work out the correct wait time
            if self.timestamps:
                if recorded_start_time is None:
                    recorded_start_time = message.timestamp

                t_wakeup = playback_start_time + (
                    message.timestamp - t_skipped - recorded_start_time
                )
            else:
                t_wakeup += self.gap

            sleep_period = t_wakeup - time.perf_counter()

            if self.skip and sleep_period > self.skip:
                t_skipped += sleep_period - self.skip
                sleep_period = self.skip

            if sleep_period > 1e-4:
                time.sleep(sleep_period)

            yield message



#!/usr/bin/env python

import unittest
from unittest.mock import Mock, patch

import can
from can.interfaces.systec import ucan, ucanbus
from can.interfaces.systec.ucan import *


class SystecTest(unittest.TestCase):
    def compare_message(self, first, second, msg):
        if (
            first.arbitration_id != second.arbitration_id
            or first.data != second.data
            or first.is_extended_id != second.is_extended_id
        ):
            raise self.failureException(msg)

    def setUp(self):
        # add equality function for can.Message
        self.addTypeEqualityFunc(can.Message, self.compare_message)

        ucan.UcanInitHwConnectControlEx = Mock()
        ucan.UcanInitHardwareEx = Mock()
        ucan.UcanInitHardwareEx2 = Mock()
        ucan.UcanInitCanEx2 = Mock()
        ucan.UcanGetHardwareInfoEx2 = Mock()
        ucan.UcanSetAcceptanceEx = Mock()
        ucan.UcanDeinitCanEx = Mock()
        ucan.UcanDeinitHardware = Mock()
        ucan.UcanWriteCanMsgEx = Mock()
        ucan.UcanResetCanEx = Mock()
        ucan._UCAN_INITIALIZED = True  # Fake this
        self.bus = can.Bus(interface="systec", channel=0, bitrate=125000)

    def test_bus_creation(self):
        self.assertIsInstance(self.bus, ucanbus.UcanBus)
        self.assertTrue(ucan.UcanInitHwConnectControlEx.called)
        self.assertTrue(
            ucan.UcanInitHardwareEx.called or ucan.UcanInitHardwareEx2.called
        )
        self.assertTrue(ucan.UcanInitCanEx2.called)
        self.assertTrue(ucan.UcanGetHardwareInfoEx2.called)
        self.assertTrue(ucan.UcanSetAcceptanceEx.called)

    def test_bus_shutdown(self):
        self.bus.shutdown()
        self.assertTrue(ucan.UcanDeinitCanEx.called)
        self.assertTrue(ucan.UcanDeinitHardware.called)

    def test_filter_setup(self):
        # no filter in the constructor
        expected_args = ((self.bus._ucan._handle, 0, AMR_ALL, ACR_ALL),)
        self.assertEqual(ucan.UcanSetAcceptanceEx.call_args, expected_args)

        # one filter is handled by the driver
        ucan.UcanSetAcceptanceEx.reset_mock()
        can_filter = (True, 0x123, 0x123, False, False)
        self.bus.set_filters(ucanbus.UcanBus.create_filter(*can_filter))
        expected_args = (
            (
                self.bus._ucan._handle,
                0,
                ucan.UcanServer.calculate_amr(*can_filter),
                ucan.UcanServer.calculate_acr(*can_filter),
            ),
        )
        self.assertEqual(ucan.UcanSetAcceptanceEx.call_args, expected_args)

        # multiple filters are handled by the bus
        ucan.UcanSetAcceptanceEx.reset_mock()
        can_filter = ((False, 0x8, 0x8, False, False), (False, 0x9, 0x9, False, False))
        self.bus.set_filters(
            ucanbus.UcanBus.create_filter(*can_filter[0])
            + ucanbus.UcanBus.create_filter(*can_filter[1])
        )
        expected_args = ((self.bus._ucan._handle, 0, AMR_ALL, ACR_ALL),)
        self.assertEqual(ucan.UcanSetAcceptanceEx.call_args, expected_args)

    @patch("can.interfaces.systec.ucan.UcanServer.write_can_msg")
    def test_send_extended(self, mock_write_can_msg):
        msg = can.Message(
            arbitration_id=0xC0FFEE, data=[0, 25, 0, 1, 3, 1, 4], is_extended_id=True
        )
        self.bus.send(msg)

        expected_args = (
            (0, [CanMsg(msg.arbitration_id, MsgFrameFormat.MSG_FF_EXT, msg.data)]),
        )
        self.assertEqual(mock_write_can_msg.call_args, expected_args)

    @patch("can.interfaces.systec.ucan.UcanServer.write_can_msg")
    def test_send_standard(self, mock_write_can_msg):
        msg = can.Message(arbitration_id=0x321, data=[50, 51], is_extended_id=False)
        self.bus.send(msg)

        expected_args = (
            (0, [CanMsg(msg.arbitration_id, MsgFrameFormat.MSG_FF_STD, msg.data)]),
        )
        self.assertEqual(mock_write_can_msg.call_args, expected_args)

    @patch("can.interfaces.systec.ucan.UcanServer.get_msg_pending")
    def test_recv_no_message(self, mock_get_msg_pending):
        mock_get_msg_pending.return_value = 0
        self.assertEqual(self.bus.recv(timeout=0.5), None)

    @patch("can.interfaces.systec.ucan.UcanServer.get_msg_pending")
    @patch("can.interfaces.systec.ucan.UcanServer.read_can_msg")
    def test_recv_extended(self, mock_read_can_msg, mock_get_msg_pending):
        mock_read_can_msg.return_value = (
            [CanMsg(0xC0FFEF, MsgFrameFormat.MSG_FF_EXT, [1, 2, 3, 4, 5, 6, 7, 8])],
            0,
        )
        mock_get_msg_pending.return_value = 1

        msg = can.Message(
            arbitration_id=0xC0FFEF, data=[1, 2, 3, 4, 5, 6, 7, 8], is_extended_id=True
        )
        can_msg = self.bus.recv()
        self.assertEqual(can_msg, msg)

    @patch("can.interfaces.systec.ucan.UcanServer.get_msg_pending")
    @patch("can.interfaces.systec.ucan.UcanServer.read_can_msg")
    def test_recv_standard(self, mock_read_can_msg, mock_get_msg_pending):
        mock_read_can_msg.return_value = (
            [CanMsg(0x321, MsgFrameFormat.MSG_FF_STD, [50, 51])],
            0,
        )
        mock_get_msg_pending.return_value = 1

        msg = can.Message(arbitration_id=0x321, data=[50, 51], is_extended_id=False)
        can_msg = self.bus.recv()
        self.assertEqual(can_msg, msg)

    @staticmethod
    def test_bus_defaults():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=0)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            0,
            InitCanParam(
                Mode.MODE_NORMAL,
                Baudrate.BAUD_500kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                DEFAULT_BUFFER_ENTRIES,
                DEFAULT_BUFFER_ENTRIES,
            ),
        )

    @staticmethod
    def test_bus_channel():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=1)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            1,
            InitCanParam(
                Mode.MODE_NORMAL,
                Baudrate.BAUD_500kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                DEFAULT_BUFFER_ENTRIES,
                DEFAULT_BUFFER_ENTRIES,
            ),
        )

    @staticmethod
    def test_bus_bitrate():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=0, bitrate=125000)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            0,
            InitCanParam(
                Mode.MODE_NORMAL,
                Baudrate.BAUD_125kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                DEFAULT_BUFFER_ENTRIES,
                DEFAULT_BUFFER_ENTRIES,
            ),
        )

    def test_bus_custom_bitrate(self):
        with self.assertRaises(ValueError):
            can.Bus(interface="systec", channel=0, bitrate=123456)

    @staticmethod
    def test_receive_own_messages():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=0, receive_own_messages=True)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            0,
            InitCanParam(
                Mode.MODE_TX_ECHO,
                Baudrate.BAUD_500kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                DEFAULT_BUFFER_ENTRIES,
                DEFAULT_BUFFER_ENTRIES,
            ),
        )

    @staticmethod
    def test_bus_passive_state():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=0, state=can.BusState.PASSIVE)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            0,
            InitCanParam(
                Mode.MODE_LISTEN_ONLY,
                Baudrate.BAUD_500kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                DEFAULT_BUFFER_ENTRIES,
                DEFAULT_BUFFER_ENTRIES,
            ),
        )

    @staticmethod
    def test_rx_buffer_entries():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=0, rx_buffer_entries=1024)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            0,
            InitCanParam(
                Mode.MODE_NORMAL,
                Baudrate.BAUD_500kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                1024,
                DEFAULT_BUFFER_ENTRIES,
            ),
        )

    @staticmethod
    def test_tx_buffer_entries():
        ucan.UcanInitCanEx2.reset_mock()
        bus = can.Bus(interface="systec", channel=0, tx_buffer_entries=1024)
        ucan.UcanInitCanEx2.assert_called_once_with(
            bus._ucan._handle,
            0,
            InitCanParam(
                Mode.MODE_NORMAL,
                Baudrate.BAUD_500kBit,
                OutputControl.OCR_DEFAULT,
                AMR_ALL,
                ACR_ALL,
                BaudrateEx.BAUDEX_USE_BTR01,
                DEFAULT_BUFFER_ENTRIES,
                1024,
            ),
        )

    def test_flush_tx_buffer(self):
        self.bus.flush_tx_buffer()
        ucan.UcanResetCanEx.assert_called_once_with(
            self.bus._ucan._handle, 0, ResetFlags.RESET_ONLY_TX_BUFF
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import asyncio
import time
import unittest

import can


class NotifierTest(unittest.TestCase):
    def test_single_bus(self):
        with can.Bus("test", interface="virtual", receive_own_messages=True) as bus:
            reader = can.BufferedReader()
            notifier = can.Notifier(bus, [reader], 0.1)
            msg = can.Message()
            bus.send(msg)
            self.assertIsNotNone(reader.get_message(1))
            notifier.stop()

    def test_multiple_bus(self):
        with can.Bus(0, interface="virtual", receive_own_messages=True) as bus1:
            with can.Bus(1, interface="virtual", receive_own_messages=True) as bus2:
                reader = can.BufferedReader()
                notifier = can.Notifier([bus1, bus2], [reader], 0.1)
                msg = can.Message()
                bus1.send(msg)
                time.sleep(0.1)
                bus2.send(msg)
                recv_msg = reader.get_message(1)
                self.assertIsNotNone(recv_msg)
                self.assertEqual(recv_msg.channel, 0)
                recv_msg = reader.get_message(1)
                self.assertIsNotNone(recv_msg)
                self.assertEqual(recv_msg.channel, 1)
                notifier.stop()


class AsyncNotifierTest(unittest.TestCase):
    def test_asyncio_notifier(self):
        async def run_it():
            with can.Bus("test", interface="virtual", receive_own_messages=True) as bus:
                reader = can.AsyncBufferedReader()
                notifier = can.Notifier(
                    bus, [reader], 0.1, loop=asyncio.get_running_loop()
                )
                bus.send(can.Message())
                recv_msg = await asyncio.wait_for(reader.get_message(), 0.5)
                self.assertIsNotNone(recv_msg)
                notifier.stop()

        asyncio.run(run_it())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
This module contains a helper for writing test cases that need to compare messages.
"""

from copy import copy


class ComparingMessagesTestCase:
    """
    Must be extended by a class also extending a unittest.TestCase.

    .. note:: This class does not extend unittest.TestCase so it does not get
              run as a test itself.
    """

    def __init__(self, allowed_timestamp_delta=0.0, preserves_channel=True):
        """
        :param float or int or None allowed_timestamp_delta: directly passed to :meth:`can.Message.equals`
        :param bool preserves_channel: if True, checks that the channel attribute is preserved
        """
        self.allowed_timestamp_delta = allowed_timestamp_delta
        self.preserves_channel = preserves_channel

    def assertMessageEqual(self, message_1, message_2):
        """
        Checks that two messages are equal, according to the given rules.
        """

        if message_1.equals(message_2, timestamp_delta=self.allowed_timestamp_delta):
            return
        elif self.preserves_channel:
            print(f"Comparing: message 1: {message_1!r}")
            print(f"           message 2: {message_2!r}")
            self.fail(
                "messages are unequal with allowed timestamp delta {}".format(
                    self.allowed_timestamp_delta
                )
            )
        else:
            message_2 = copy(message_2)  # make sure this method is pure
            message_2.channel = message_1.channel
            if message_1.equals(
                message_2, timestamp_delta=self.allowed_timestamp_delta
            ):
                return
            else:
                print(f"Comparing: message 1: {message_1!r}")
                print(f"           message 2: {message_2!r}")
                self.fail(
                    "messages are unequal with allowed timestamp delta {} even when ignoring channels".format(
                        self.allowed_timestamp_delta
                    )
                )

    def assertMessagesEqual(self, messages_1, messages_2):
        """
        Checks the order and content of the individual messages pairwise.
        Raises an error if the lengths of the sequences are not equal.
        """
        self.assertEqual(
            len(messages_1), len(messages_2), "the number of messages differs"
        )

        for message_1, message_2 in zip(messages_1, messages_2):
            self.assertMessageEqual(message_1, message_2)



"""
The ``can`` package provides controller area network support for
Python developers; providing common abstractions to
different hardware devices, and a suite of utilities for sending and receiving
messages on a can bus.
"""

import logging
from typing import Any, Dict

__version__ = "4.2.0rc0"
__all__ = [
    "ASCReader",
    "ASCWriter",
    "AsyncBufferedReader",
    "BitTiming",
    "BitTimingFd",
    "BLFReader",
    "BLFWriter",
    "BufferedReader",
    "Bus",
    "BusABC",
    "BusState",
    "CanError",
    "CanInitializationError",
    "CanInterfaceNotImplementedError",
    "CanOperationError",
    "CanTimeoutError",
    "CanutilsLogReader",
    "CanutilsLogWriter",
    "CSVReader",
    "CSVWriter",
    "CyclicSendTaskABC",
    "LimitedDurationCyclicSendTaskABC",
    "Listener",
    "Logger",
    "LogReader",
    "ModifiableCyclicTaskABC",
    "Message",
    "MessageSync",
    "MF4Reader",
    "MF4Writer",
    "Notifier",
    "Printer",
    "RedirectReader",
    "RestartableCyclicTaskABC",
    "SizedRotatingLogger",
    "SqliteReader",
    "SqliteWriter",
    "ThreadSafeBus",
    "TRCFileVersion",
    "TRCReader",
    "TRCWriter",
    "VALID_INTERFACES",
    "bit_timing",
    "broadcastmanager",
    "bus",
    "ctypesutil",
    "detect_available_configs",
    "exceptions",
    "interface",
    "interfaces",
    "listener",
    "logconvert",
    "log",
    "logger",
    "message",
    "notifier",
    "player",
    "set_logging_level",
    "thread_safe_bus",
    "typechecking",
    "util",
    "viewer",
]

log = logging.getLogger("can")

rc: Dict[str, Any] = {}

from . import typechecking  # isort:skip
from . import util  # isort:skip
from . import broadcastmanager, interface
from .bit_timing import BitTiming, BitTimingFd
from .broadcastmanager import (
    CyclicSendTaskABC,
    LimitedDurationCyclicSendTaskABC,
    ModifiableCyclicTaskABC,
    RestartableCyclicTaskABC,
)
from .bus import BusABC, BusState
from .exceptions import (
    CanError,
    CanInitializationError,
    CanInterfaceNotImplementedError,
    CanOperationError,
    CanTimeoutError,
)
from .interface import Bus, detect_available_configs
from .interfaces import VALID_INTERFACES
from .io import (
    ASCReader,
    ASCWriter,
    BLFReader,
    BLFWriter,
    CanutilsLogReader,
    CanutilsLogWriter,
    CSVReader,
    CSVWriter,
    Logger,
    LogReader,
    MessageSync,
    MF4Reader,
    MF4Writer,
    Printer,
    SizedRotatingLogger,
    SqliteReader,
    SqliteWriter,
    TRCFileVersion,
    TRCReader,
    TRCWriter,
)
from .listener import AsyncBufferedReader, BufferedReader, Listener, RedirectReader
from .message import Message
from .notifier import Notifier
from .thread_safe_bus import ThreadSafeBus
from .util import set_logging_level



import logging
import time
from collections import deque
from ctypes import c_ubyte
from typing import Any, Deque, Dict, Optional, Sequence, Tuple, Union

import canalystii as driver

from can import BitTiming, BitTimingFd, BusABC, Message
from can.exceptions import CanTimeoutError
from can.typechecking import CanFilters
from can.util import check_or_adjust_timing_clock, deprecated_args_alias

logger = logging.getLogger(__name__)


class CANalystIIBus(BusABC):
    @deprecated_args_alias(
        deprecation_start="4.2.0", deprecation_end="5.0.0", bit_timing="timing"
    )
    def __init__(
        self,
        channel: Union[int, Sequence[int], str] = (0, 1),
        device: int = 0,
        bitrate: Optional[int] = None,
        timing: Optional[Union[BitTiming, BitTimingFd]] = None,
        can_filters: Optional[CanFilters] = None,
        rx_queue_size: Optional[int] = None,
        **kwargs: Dict[str, Any],
    ):
        """

        :param channel:
            Optional channel number, list/tuple of multiple channels, or comma
            separated string of channels. Default is to configure both
            channels.
        :param device:
            Optional USB device number. Default is 0 (first device found).
        :param bitrate:
            CAN bitrate in bits/second. Required unless the bit_timing argument is set.
        :param timing:
            Optional :class:`~can.BitTiming` instance to use for custom bit timing setting.
            If this argument is set then it overrides the bitrate argument. The
            `f_clock` value of the timing instance must be set to 8_000_000 (8MHz)
            for standard CAN.
            CAN FD and the :class:`~can.BitTimingFd` class are not supported.
        :param can_filters:
            Optional filters for received CAN messages.
        :param rx_queue_size:
            If set, software received message queue can only grow to this many
            messages (for all channels) before older messages are dropped
        """
        if not (bitrate or timing):
            raise ValueError("Either bitrate or timing argument is required")

        # Do this after the error handling
        super().__init__(channel=channel, can_filters=can_filters, **kwargs)

        if isinstance(channel, str):
            # Assume comma separated string of channels
            self.channels = [int(ch.strip()) for ch in channel.split(",")]
        elif isinstance(channel, int):
            self.channels = [channel]
        else:  # Sequence[int]
            self.channels = list(channel)

        self.rx_queue: Deque[Tuple[int, driver.Message]] = deque(maxlen=rx_queue_size)

        self.channel_info = f"CANalyst-II: device {device}, channels {self.channels}"

        self.device = driver.CanalystDevice(device_index=device)
        for single_channel in self.channels:
            if isinstance(timing, BitTiming):
                timing = check_or_adjust_timing_clock(timing, valid_clocks=[8_000_000])
                self.device.init(
                    single_channel, timing0=timing.btr0, timing1=timing.btr1
                )
            elif isinstance(timing, BitTimingFd):
                raise NotImplementedError(
                    f"CAN FD is not supported by {self.__class__.__name__}."
                )
            else:
                self.device.init(single_channel, bitrate=bitrate)

    # Delay to use between each poll for new messages
    #
    # The timeout is deliberately kept low to avoid the possibility of
    # a hardware buffer overflow. This value was determined
    # experimentally, but the ideal value will depend on the specific
    # system.
    RX_POLL_DELAY = 0.020

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        """Send a CAN message to the bus

        :param msg: message to send
        :param timeout: timeout (in seconds) to wait for the TX queue to clear.
        If set to ``None`` (default) the function returns immediately.

        Note: Due to limitations in the device firmware and protocol, the
        timeout will not trigger if there are problems with CAN arbitration,
        but only if the device is overloaded with a backlog of too many
        messages to send.
        """
        raw_message = driver.Message(
            msg.arbitration_id,
            0,  # timestamp
            1,  # time_flag
            0,  # send_type
            msg.is_remote_frame,
            msg.is_extended_id,
            msg.dlc,
            (c_ubyte * 8)(*msg.data),
        )

        if msg.channel is not None:
            channel = msg.channel
        elif len(self.channels) == 1:
            channel = self.channels[0]
        else:
            raise ValueError(
                "Message channel must be set when using multiple channels."
            )

        send_result = self.device.send(channel, [raw_message], timeout)
        if timeout is not None and not send_result:
            raise CanTimeoutError(f"Send timed out after {timeout} seconds")

    def _recv_from_queue(self) -> Tuple[Message, bool]:
        """Return a message from the internal receive queue"""
        channel, raw_msg = self.rx_queue.popleft()

        # Protocol timestamps are in units of 100us, convert to seconds as
        # float
        timestamp = raw_msg.timestamp * 100e-6

        return (
            Message(
                channel=channel,
                timestamp=timestamp,
                arbitration_id=raw_msg.can_id,
                is_extended_id=raw_msg.extended,
                is_remote_frame=raw_msg.remote,
                dlc=raw_msg.data_len,
                data=bytes(raw_msg.data),
            ),
            False,
        )

    def poll_received_messages(self) -> None:
        """Poll new messages from the device into the rx queue but don't
        return any message to the caller

        Calling this function isn't necessary as polling the device is done
        automatically when calling recv(). This function is for the situation
        where an application needs to empty the hardware receive buffer without
        consuming any message.
        """
        for channel in self.channels:
            self.rx_queue.extend(
                (channel, raw_msg) for raw_msg in self.device.receive(channel)
            )

    def _recv_internal(
        self, timeout: Optional[float] = None
    ) -> Tuple[Optional[Message], bool]:
        """

        :param timeout: float in seconds
        :return:
        """

        if self.rx_queue:
            return self._recv_from_queue()

        deadline = None
        while deadline is None or time.time() < deadline:
            if deadline is None and timeout is not None:
                deadline = time.time() + timeout

            self.poll_received_messages()

            if self.rx_queue:
                return self._recv_from_queue()

            # If blocking on a timeout, add a sleep before we loop again
            # to reduce CPU usage.
            if deadline is None or deadline - time.time() > 0.050:
                time.sleep(self.RX_POLL_DELAY)

        return (None, False)

    def flush_tx_buffer(self, channel: Optional[int] = None) -> None:
        """Flush the TX buffer of the device.

        :param channel:
            Optional channel number to flush. If set to None, all initialized
            channels are flushed.

        Note that because of protocol limitations this function returning
        doesn't mean that messages have been sent, it may also mean they
        failed to send.
        """
        if channel:
            self.device.flush_tx_buffer(channel, float("infinity"))
        else:
            for ch in self.channels:
                self.device.flush_tx_buffer(ch, float("infinity"))

    def shutdown(self) -> None:
        super().shutdown()
        for channel in self.channels:
            self.device.stop(channel)
        self.device = None



#!/usr/bin/env python

"""
This module tests the functions inside of player.py
"""

import io
import os
import sys
import unittest
from unittest import mock
from unittest.mock import Mock

import can
import can.player


class TestPlayerScriptModule(unittest.TestCase):
    logfile = os.path.join(os.path.dirname(__file__), "data", "test_CanMessage.asc")

    def setUp(self) -> None:
        # Patch VirtualBus object
        patcher_virtual_bus = mock.patch("can.interfaces.virtual.VirtualBus", spec=True)
        self.MockVirtualBus = patcher_virtual_bus.start()
        self.addCleanup(patcher_virtual_bus.stop)
        self.mock_virtual_bus = self.MockVirtualBus.return_value
        self.mock_virtual_bus.__enter__ = Mock(return_value=self.mock_virtual_bus)

        # Patch time sleep object
        patcher_sleep = mock.patch("can.io.player.time.sleep", spec=True)
        self.MockSleep = patcher_sleep.start()
        self.addCleanup(patcher_sleep.stop)

        self.baseargs = [sys.argv[0], "-i", "virtual"]

    def assertSuccessfulCleanup(self):
        self.MockVirtualBus.assert_called_once()
        self.mock_virtual_bus.__exit__.assert_called_once()

    def test_play_virtual(self):
        sys.argv = self.baseargs + [self.logfile]
        can.player.main()
        msg1 = can.Message(
            timestamp=2.501,
            arbitration_id=0xC8,
            is_extended_id=False,
            is_fd=False,
            is_rx=False,
            channel=1,
            dlc=8,
            data=[0x9, 0x8, 0x7, 0x6, 0x5, 0x4, 0x3, 0x2],
        )
        msg2 = can.Message(
            timestamp=17.876708,
            arbitration_id=0x6F9,
            is_extended_id=False,
            is_fd=False,
            is_rx=True,
            channel=0,
            dlc=8,
            data=[0x5, 0xC, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0],
        )
        if sys.version_info >= (3, 8):
            # The args argument was introduced with python 3.8
            self.assertTrue(
                msg1.equals(self.mock_virtual_bus.send.mock_calls[0].args[0])
            )
            self.assertTrue(
                msg2.equals(self.mock_virtual_bus.send.mock_calls[1].args[0])
            )
        self.assertSuccessfulCleanup()

    def test_play_virtual_verbose(self):
        sys.argv = self.baseargs + ["-v", self.logfile]
        with unittest.mock.patch("sys.stdout", new_callable=io.StringIO) as mock_stdout:
            can.player.main()
        self.assertIn("09 08 07 06 05 04 03 02", mock_stdout.getvalue())
        self.assertIn("05 0c 00 00 00 00 00 00", mock_stdout.getvalue())
        self.assertEqual(self.mock_virtual_bus.send.call_count, 2)
        self.assertSuccessfulCleanup()

    def test_play_virtual_exit(self):
        self.MockSleep.side_effect = [None, KeyboardInterrupt]

        sys.argv = self.baseargs + [self.logfile]
        can.player.main()
        assert self.mock_virtual_bus.send.call_count <= 2
        self.assertSuccessfulCleanup()

    def test_play_skip_error_frame(self):
        logfile = os.path.join(
            os.path.dirname(__file__), "data", "logfile_errorframes.asc"
        )
        sys.argv = self.baseargs + ["-v", logfile]
        can.player.main()
        self.assertEqual(self.mock_virtual_bus.send.call_count, 9)
        self.assertSuccessfulCleanup()

    def test_play_error_frame(self):
        logfile = os.path.join(
            os.path.dirname(__file__), "data", "logfile_errorframes.asc"
        )
        sys.argv = self.baseargs + ["-v", "--error-frames", logfile]
        can.player.main()
        self.assertEqual(self.mock_virtual_bus.send.call_count, 12)
        self.assertSuccessfulCleanup()


class TestPlayerCompressedFile(TestPlayerScriptModule):
    """
    Re-run tests using a compressed file.
    """

    logfile = os.path.join(os.path.dirname(__file__), "data", "test_CanMessage.asc.gz")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest
import warnings

import pytest

from can import BitTiming, BitTimingFd
from can.exceptions import CanInitializationError
from can.util import (
    _create_bus_config,
    _rename_kwargs,
    cast_from_string,
    channel2int,
    check_or_adjust_timing_clock,
    deprecated_args_alias,
)


class RenameKwargsTest(unittest.TestCase):
    expected_kwargs = dict(a=1, b=2, c=3, d=4)

    def _test(self, start: str, end: str, kwargs, aliases):
        # Test that we do get the DeprecationWarning when called with deprecated kwargs
        with self.assertWarnsRegex(
            DeprecationWarning, "is deprecated.*?" + start + ".*?" + end
        ):
            _rename_kwargs("unit_test", start, end, kwargs, aliases)

        # Test that the aliases contains the deprecated values and
        # the obsolete kwargs have been removed
        assert kwargs == self.expected_kwargs

        # Test that we do not get a DeprecationWarning when we call
        # without deprecated kwargs

        # Cause all warnings to always be triggered.
        warnings.simplefilter("error", DeprecationWarning)
        try:
            _rename_kwargs("unit_test", start, end, kwargs, aliases)
        finally:
            warnings.resetwarnings()

    def test_rename(self):
        kwargs = dict(old_a=1, old_b=2, c=3, d=4)
        aliases = {"old_a": "a", "old_b": "b"}
        self._test("1.0", "2.0", kwargs, aliases)

    def test_obsolete(self):
        kwargs = dict(a=1, b=2, c=3, d=4, z=10)
        aliases = {"z": None}
        self._test("1.0", "2.0", kwargs, aliases)

    def test_rename_and_obsolete(self):
        kwargs = dict(old_a=1, old_b=2, c=3, d=4, z=10)
        aliases = {"old_a": "a", "old_b": "b", "z": None}
        self._test("1.0", "2.0", kwargs, aliases)

    def test_with_new_and_alias_present(self):
        kwargs = dict(old_a=1, a=1, b=2, c=3, d=4, z=10)
        aliases = {"old_a": "a", "old_b": "b", "z": None}
        with self.assertRaises(TypeError):
            self._test("1.0", "2.0", kwargs, aliases)


class DeprecatedArgsAliasTest(unittest.TestCase):
    def test_decorator(self):
        @deprecated_args_alias("1.0.0", old_a="a")
        def _test_func1(a):
            pass

        with pytest.warns(DeprecationWarning) as record:
            _test_func1(old_a=1)
            assert len(record) == 1
            assert (
                record[0].message.args[0]
                == "The 'old_a' argument is deprecated since python-can v1.0.0. Use 'a' instead."
            )

        @deprecated_args_alias("1.6.0", "3.4.0", old_a="a", old_b=None)
        def _test_func2(a):
            pass

        with pytest.warns(DeprecationWarning) as record:
            _test_func2(old_a=1, old_b=2)
            assert len(record) == 2
            assert record[0].message.args[0] == (
                "The 'old_a' argument is deprecated since python-can v1.6.0, and scheduled for "
                "removal in python-can v3.4.0. Use 'a' instead."
            )
            assert record[1].message.args[0] == (
                "The 'old_b' argument is deprecated since python-can v1.6.0, and scheduled for "
                "removal in python-can v3.4.0."
            )

        @deprecated_args_alias("1.6.0", "3.4.0", old_a="a")
        @deprecated_args_alias("2.0.0", "4.0.0", old_b=None)
        def _test_func3(a):
            pass

        with pytest.warns(DeprecationWarning) as record:
            _test_func3(old_a=1, old_b=2)
            assert len(record) == 2
            assert record[0].message.args[0] == (
                "The 'old_a' argument is deprecated since python-can v1.6.0, and scheduled "
                "for removal in python-can v3.4.0. Use 'a' instead."
            )
            assert record[1].message.args[0] == (
                "The 'old_b' argument is deprecated since python-can v2.0.0, and scheduled "
                "for removal in python-can v4.0.0."
            )

        with pytest.warns(DeprecationWarning) as record:
            _test_func3(old_a=1)
            assert len(record) == 1
            assert record[0].message.args[0] == (
                "The 'old_a' argument is deprecated since python-can v1.6.0, and scheduled "
                "for removal in python-can v3.4.0. Use 'a' instead."
            )

        with pytest.warns(DeprecationWarning) as record:
            _test_func3(a=1, old_b=2)
            assert len(record) == 1
            assert record[0].message.args[0] == (
                "The 'old_b' argument is deprecated since python-can v2.0.0, and scheduled "
                "for removal in python-can v4.0.0."
            )

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            _test_func3(a=1)


class TestBusConfig(unittest.TestCase):
    base_config = dict(interface="socketcan", bitrate=500_000)
    port_alpha_config = dict(interface="socketcan", bitrate=500_000, port="fail123")
    port_to_high_config = dict(interface="socketcan", bitrate=500_000, port="999999")
    port_wrong_type_config = dict(interface="socketcan", bitrate=500_000, port=(1234,))

    def test_timing_can_use_int(self):
        """
        Test that an exception is not raised when using
        integers for timing values in config.
        """
        timing_conf = dict(tseg1=5, tseg2=10, sjw=25)
        try:
            _create_bus_config({**self.base_config, **timing_conf})
        except TypeError as e:
            self.fail(e)
        self.assertRaises(
            ValueError, _create_bus_config, {**self.port_alpha_config, **timing_conf}
        )
        self.assertRaises(
            ValueError, _create_bus_config, {**self.port_to_high_config, **timing_conf}
        )
        self.assertRaises(
            TypeError,
            _create_bus_config,
            {**self.port_wrong_type_config, **timing_conf},
        )


class TestChannel2Int(unittest.TestCase):
    def test_channel2int(self) -> None:
        self.assertEqual(0, channel2int("can0"))
        self.assertEqual(0, channel2int("vcan0"))
        self.assertEqual(1, channel2int("vcan1"))
        self.assertEqual(12, channel2int("vcan12"))
        self.assertEqual(3, channel2int(3))
        self.assertEqual(42, channel2int("42"))
        self.assertEqual(None, channel2int("can"))
        self.assertEqual(None, channel2int("can0a"))


class TestCheckAdjustTimingClock(unittest.TestCase):
    def test_adjust_timing(self):
        timing = BitTiming(f_clock=80_000_000, brp=10, tseg1=13, tseg2=2, sjw=1)

        # Check identity case
        new_timing = check_or_adjust_timing_clock(timing, valid_clocks=[80_000_000])
        assert timing == new_timing

        with pytest.warns(UserWarning) as record:
            new_timing = check_or_adjust_timing_clock(
                timing, valid_clocks=[8_000_000, 24_000_000]
            )
            assert len(record) == 1
            assert (
                record[0].message.args[0]
                == "Adjusted f_clock in BitTiming from 80000000 to 8000000"
            )
            assert new_timing.__class__ == BitTiming
            assert new_timing.f_clock == 8_000_000
            assert new_timing.bitrate == timing.bitrate
            assert new_timing.tseg1 == timing.tseg1
            assert new_timing.tseg2 == timing.tseg2
            assert new_timing.sjw == timing.sjw

        # Check that order is preserved
        with pytest.warns(UserWarning) as record:
            new_timing = check_or_adjust_timing_clock(
                timing, valid_clocks=[24_000_000, 8_000_000]
            )
            assert new_timing.f_clock == 24_000_000
            assert len(record) == 1
            assert (
                record[0].message.args[0]
                == "Adjusted f_clock in BitTiming from 80000000 to 24000000"
            )

        # Check that order is preserved for all valid clock rates
        with pytest.warns(UserWarning) as record:
            new_timing = check_or_adjust_timing_clock(
                timing, valid_clocks=[8_000, 24_000_000, 8_000_000]
            )
            assert new_timing.f_clock == 24_000_000
            assert len(record) == 1
            assert (
                record[0].message.args[0]
                == "Adjusted f_clock in BitTiming from 80000000 to 24000000"
            )

        with pytest.raises(CanInitializationError):
            check_or_adjust_timing_clock(timing, valid_clocks=[8_000, 16_000])

    def test_adjust_timing_fd(self):
        timing = BitTimingFd(
            f_clock=160_000_000,
            nom_brp=2,
            nom_tseg1=119,
            nom_tseg2=40,
            nom_sjw=40,
            data_brp=2,
            data_tseg1=29,
            data_tseg2=10,
            data_sjw=10,
        )

        # Check identity case
        new_timing = check_or_adjust_timing_clock(timing, valid_clocks=[160_000_000])
        assert timing == new_timing

        with pytest.warns(UserWarning) as record:
            new_timing = check_or_adjust_timing_clock(
                timing, valid_clocks=[8_000, 80_000_000]
            )
            assert len(record) == 1, "; ".join(
                [record[i].message.args[0] for i in range(len(record))]
            )  # print all warnings, if more than one warning is present
            assert (
                record[0].message.args[0]
                == "Adjusted f_clock in BitTimingFd from 160000000 to 80000000"
            )
            assert new_timing.__class__ == BitTimingFd
            assert new_timing.f_clock == 80_000_000
            assert new_timing.nom_bitrate == 500_000
            assert new_timing.nom_tseg1 == 119
            assert new_timing.nom_tseg2 == 40
            assert new_timing.nom_sjw == 40
            assert new_timing.data_bitrate == 2_000_000
            assert new_timing.data_tseg1 == 29
            assert new_timing.data_tseg2 == 10
            assert new_timing.data_sjw == 10

        with pytest.raises(CanInitializationError):
            check_or_adjust_timing_clock(timing, valid_clocks=[8_000, 16_000])


class TestCastFromString(unittest.TestCase):
    def test_cast_from_string(self) -> None:
        self.assertEqual(1, cast_from_string("1"))
        self.assertEqual(-1, cast_from_string("-1"))
        self.assertEqual(0, cast_from_string("-0"))
        self.assertEqual(1.1, cast_from_string("1.1"))
        self.assertEqual(-1.1, cast_from_string("-1.1"))
        self.assertEqual(0.1, cast_from_string(".1"))
        self.assertEqual(10.0, cast_from_string(".1e2"))
        self.assertEqual(0.001, cast_from_string(".1e-2"))
        self.assertEqual(-0.001, cast_from_string("-.1e-2"))
        self.assertEqual("text", cast_from_string("text"))
        self.assertEqual("", cast_from_string(""))
        self.assertEqual("can0", cast_from_string("can0"))
        self.assertEqual("0can", cast_from_string("0can"))
        self.assertEqual(False, cast_from_string("false"))
        self.assertEqual(False, cast_from_string("False"))
        self.assertEqual(True, cast_from_string("true"))
        self.assertEqual(True, cast_from_string("True"))

        with self.assertRaises(TypeError):
            cast_from_string(None)



#!/usr/bin/env python

"""
Tests helpers in `can.interfaces.socketcan.socketcan_common`.
"""

import gzip
import unittest
from base64 import b64decode
from unittest import mock

from can.interfaces.socketcan.utils import error_code_to_str, find_available_interfaces

from .config import IS_LINUX, TEST_INTERFACE_SOCKETCAN


class TestSocketCanHelpers(unittest.TestCase):
    @unittest.skipUnless(IS_LINUX, "socketcan is only available on Linux")
    def test_error_code_to_str(self):
        """
        Check that the function does not crash & always
        returns at least one character.
        """

        # all possible & also some invalid error codes
        test_data = list(range(0, 256)) + [-1, 256, 5235, 346264, None]

        for error_code in test_data:
            string = error_code_to_str(error_code)
            self.assertTrue(string)  # not None or empty

    @unittest.skipUnless(
        TEST_INTERFACE_SOCKETCAN, "socketcan is only available on Linux"
    )
    def test_find_available_interfaces(self):
        result = find_available_interfaces()

        self.assertGreaterEqual(len(result), 3)
        self.assertIn("vcan0", result)
        self.assertIn("vxcan0", result)
        self.assertIn("slcan0", result)

    def test_find_available_interfaces_w_patch(self):
        # Contains lo, eth0, wlan0, vcan0, mycustomCan123
        ip_output_gz_b64 = (
            "H4sIAAAAAAAAA+2UzW+CMBjG7/wVhrNL+BC29IboEqNSwzQejDEViiMC5aNsmmX/+wpZTGUwDAcP"
            "y5qmh+d5++bN80u7EXpsfZRnsUTf8yMXn0TQk/u8GqEQM1EMiMjpXoAOGZM3F6mUZxAuhoY55UpL"
            "fbWoKjO4Hts7pl/kLdc+pDlrrmuaqnNq4vqZU8wSkSTHOeYHIjFOM4poOevKmlpwbfF+4EfHkLil"
            "PRo/G6vZkrcPKcnjwnOxh/KA8h49JQGOimAkSaq03NFz/B0PiffIOfIXkeumOCtiEiUJXG++bp8S"
            "5Dooo/WVZeFnvxmYUgsM01fpBmQWfDAN256M7SqioQ2NkWm8LKvGnIU3qTN+xylrV/FdaHrJzmFk"
            "gkacozuzZMnhtAGkLANFAaoKBgOgaUDXG0F6Hrje7SDVWpDvAYpuIdmJV4dn2cSx9VUuGiFCe25Y"
            "fwTi4KmW4ptzG0ULGvYPLN1APSqdMN3/82TRtOeqSbW5hmcnzygJTRTJivofcEvAgrAVvgD8aLkv"
            "/AcAAA=="
        )
        ip_output = gzip.decompress(b64decode(ip_output_gz_b64)).decode("ascii")

        with mock.patch("subprocess.check_output") as check_output:
            check_output.return_value = ip_output
            ifs = find_available_interfaces()

            self.assertEqual(["vcan0", "mycustomCan123"], ifs)

    def test_find_available_interfaces_exception(self):
        with mock.patch("subprocess.check_output") as check_output:
            check_output.return_value = "<h1>Not JSON</h1>"
            result = find_available_interfaces()
            self.assertEqual([], result)

            check_output.side_effect = Exception("Something went wrong :-/")
            result = find_available_interfaces()
            self.assertEqual([], result)


if __name__ == "__main__":
    unittest.main()



"""
A text based interface. For example use over serial ports like
"/dev/ttyS1" or "/dev/ttyUSB0" on Linux machines or "COM1" on Windows.
The interface is a simple implementation that has been used for
recording CAN traces.

See the interface documentation for the format being used.
"""

import io
import logging
import struct
from typing import Any, List, Optional, Tuple

from can import (
    BusABC,
    CanInitializationError,
    CanInterfaceNotImplementedError,
    CanOperationError,
    CanTimeoutError,
    Message,
)
from can.typechecking import AutoDetectedConfig

logger = logging.getLogger("can.serial")

try:
    import serial
except ImportError:
    logger.warning(
        "You won't be able to use the serial can backend without "
        "the serial module installed!"
    )
    serial = None

try:
    from serial.tools.list_ports import comports as list_comports
except ImportError:
    # If unavailable on some platform, just return nothing
    def list_comports() -> List[Any]:
        return []


class SerialBus(BusABC):
    """
    Enable basic can communication over a serial device.

    .. note:: See :meth:`~_recv_internal` for some special semantics.

    """

    def __init__(
        self,
        channel: str,
        baudrate: int = 115200,
        timeout: float = 0.1,
        rtscts: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """
        :param channel:
            The serial device to open. For example "/dev/ttyS1" or
            "/dev/ttyUSB0" on Linux or "COM1" on Windows systems.

        :param baudrate:
            Baud rate of the serial device in bit/s (default 115200).

            .. warning::
                Some serial port implementations don't care about the baudrate.

        :param timeout:
            Timeout for the serial device in seconds (default 0.1).

        :param rtscts:
            turn hardware handshake (RTS/CTS) on and off

        :raises ~can.exceptions.CanInitializationError:
            If the given parameters are invalid.
        :raises ~can.exceptions.CanInterfaceNotImplementedError:
            If the serial module is not installed.
        """

        if not serial:
            raise CanInterfaceNotImplementedError("the serial module is not installed")

        if not channel:
            raise TypeError("Must specify a serial port.")

        self.channel_info = f"Serial interface: {channel}"

        try:
            self._ser = serial.serial_for_url(
                channel, baudrate=baudrate, timeout=timeout, rtscts=rtscts
            )
        except ValueError as error:
            raise CanInitializationError(
                "could not create the serial device"
            ) from error

        super().__init__(channel, *args, **kwargs)

    def shutdown(self) -> None:
        """
        Close the serial interface.
        """
        super().shutdown()
        self._ser.close()

    def send(self, msg: Message, timeout: Optional[float] = None) -> None:
        """
        Send a message over the serial device.

        :param msg:
            Message to send.

            .. note:: Flags like ``extended_id``, ``is_remote_frame`` and
                      ``is_error_frame`` will be ignored.

            .. note:: If the timestamp is a float value it will be converted
                      to an integer.

        :param timeout:
            This parameter will be ignored. The timeout value of the channel is
            used instead.

        """
        # Pack timestamp
        try:
            timestamp = struct.pack("<I", int(msg.timestamp * 1000))
        except struct.error:
            raise ValueError("Timestamp is out of range")

        # Pack arbitration ID
        try:
            arbitration_id = struct.pack("<I", msg.arbitration_id)
        except struct.error:
            raise ValueError("Arbitration ID is out of range")

        # Assemble message
        byte_msg = bytearray()
        byte_msg.append(0xAA)
        byte_msg += timestamp
        byte_msg.append(msg.dlc)
        byte_msg += arbitration_id
        byte_msg += msg.data
        byte_msg.append(0xBB)

        # Write to serial device
        try:
            self._ser.write(byte_msg)
        except serial.PortNotOpenError as error:
            raise CanOperationError("writing to closed port") from error
        except serial.SerialTimeoutException as error:
            raise CanTimeoutError() from error

    def _recv_internal(
        self, timeout: Optional[float]
    ) -> Tuple[Optional[Message], bool]:
        """
        Read a message from the serial device.

        :param timeout:

            .. warning::
                This parameter will be ignored. The timeout value of the channel is used.

        :returns:
            Received message and :obj:`False` (because no filtering as taken place).

            .. warning::
                Flags like ``is_extended_id``, ``is_remote_frame`` and ``is_error_frame``
                will not be set over this function, the flags in the return
                message are the default values.
        """
        try:
            rx_byte = self._ser.read()
            if rx_byte and ord(rx_byte) == 0xAA:
                s = self._ser.read(4)
                timestamp = struct.unpack("<I", s)[0]
                dlc = ord(self._ser.read())
                if dlc > 8:
                    raise ValueError("received DLC may not exceed 8 bytes")

                s = self._ser.read(4)
                arbitration_id = struct.unpack("<I", s)[0]
                if arbitration_id >= 0x20000000:
                    raise ValueError(
                        "received arbitration id may not exceed 2^29 (0x20000000)"
                    )

                data = self._ser.read(dlc)

                delimiter_byte = ord(self._ser.read())
                if delimiter_byte == 0xBB:
                    # received message data okay
                    msg = Message(
                        # TODO: We are only guessing that they are milliseconds
                        timestamp=timestamp / 1000,
                        arbitration_id=arbitration_id,
                        dlc=dlc,
                        data=data,
                    )
                    return msg, False

                else:
                    raise CanOperationError(
                        f"invalid delimiter byte while reading message: {delimiter_byte}"
                    )

            else:
                return None, False

        except serial.SerialException as error:
            raise CanOperationError("could not read from serial") from error

    def fileno(self) -> int:
        try:
            return self._ser.fileno()
        except io.UnsupportedOperation:
            raise NotImplementedError(
                "fileno is not implemented using current CAN bus on this platform"
            )
        except Exception as exception:
            raise CanOperationError("Cannot fetch fileno") from exception

    @staticmethod
    def _detect_available_configs() -> List[AutoDetectedConfig]:
        return [
            {"interface": "serial", "channel": port.device} for port in list_comports()
        ]



"""
Exposes several methods for transmitting cyclic messages.

The main entry point to these classes should be through
:meth:`can.BusABC.send_periodic`.
"""

import abc
import logging
import sys
import threading
import time
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Tuple, Union

from typing_extensions import Final

from can import typechecking
from can.message import Message

if TYPE_CHECKING:
    from can.bus import BusABC


# try to import win32event for event-based cyclic send task (needs the pywin32 package)
USE_WINDOWS_EVENTS = False
try:
    import win32event

    # Python 3.11 provides a more precise sleep implementation on Windows, so this is not necessary.
    # Put version check here, so mypy does not complain about `win32event` not being defined.
    if sys.version_info < (3, 11):
        USE_WINDOWS_EVENTS = True
except ImportError:
    pass

log = logging.getLogger("can.bcm")

NANOSECONDS_IN_SECOND: Final[int] = 1_000_000_000
NANOSECONDS_IN_MILLISECOND: Final[int] = 1_000_000


class CyclicTask(abc.ABC):
    """
    Abstract Base for all cyclic tasks.
    """

    @abc.abstractmethod
    def stop(self) -> None:
        """Cancel this periodic task.

        :raises ~can.exceptions.CanError:
            If stop is called on an already stopped task.
        """


class CyclicSendTaskABC(CyclicTask):
    """
    Message send task with defined period
    """

    def __init__(
        self, messages: Union[Sequence[Message], Message], period: float
    ) -> None:
        """
        :param messages:
            The messages to be sent periodically.
        :param period: The rate in seconds at which to send the messages.

        :raises ValueError: If the given messages are invalid
        """
        messages = self._check_and_convert_messages(messages)

        # Take the Arbitration ID of the first element
        self.arbitration_id = messages[0].arbitration_id
        self.period = period
        self.period_ns = int(round(period * 1e9))
        self.messages = messages

    @staticmethod
    def _check_and_convert_messages(
        messages: Union[Sequence[Message], Message]
    ) -> Tuple[Message, ...]:
        """Helper function to convert a Message or Sequence of messages into a
        tuple, and raises an error when the given value is invalid.

        Performs error checking to ensure that all Messages have the same
        arbitration ID and channel.

        Should be called when the cyclic task is initialized.

        :raises ValueError: If the given messages are invalid
        """
        if not isinstance(messages, (list, tuple)):
            if isinstance(messages, Message):
                messages = [messages]
            else:
                raise ValueError("Must be either a list, tuple, or a Message")
        if not messages:
            raise ValueError("Must be at least a list or tuple of length 1")
        messages = tuple(messages)

        all_same_id = all(
            message.arbitration_id == messages[0].arbitration_id for message in messages
        )
        if not all_same_id:
            raise ValueError("All Arbitration IDs should be the same")

        all_same_channel = all(
            message.channel == messages[0].channel for message in messages
        )
        if not all_same_channel:
            raise ValueError("All Channel IDs should be the same")

        return messages


class LimitedDurationCyclicSendTaskABC(CyclicSendTaskABC):
    def __init__(
        self,
        messages: Union[Sequence[Message], Message],
        period: float,
        duration: Optional[float],
    ) -> None:
        """Message send task with a defined duration and period.

        :param messages:
            The messages to be sent periodically.
        :param period: The rate in seconds at which to send the messages.
        :param duration:
            Approximate duration in seconds to continue sending messages. If
            no duration is provided, the task will continue indefinitely.

        :raises ValueError: If the given messages are invalid
        """
        super().__init__(messages, period)
        self.duration = duration


class RestartableCyclicTaskABC(CyclicSendTaskABC):
    """Adds support for restarting a stopped cyclic task"""

    @abc.abstractmethod
    def start(self) -> None:
        """Restart a stopped periodic task."""


class ModifiableCyclicTaskABC(CyclicSendTaskABC):
    """Adds support for modifying a periodic message"""

    def _check_modified_messages(self, messages: Tuple[Message, ...]) -> None:
        """Helper function to perform error checking when modifying the data in
        the cyclic task.

        Performs error checking to ensure the arbitration ID and the number of
        cyclic messages hasn't changed.

        Should be called when modify_data is called in the cyclic task.

        :raises ValueError: If the given messages are invalid
        """
        if len(self.messages) != len(messages):
            raise ValueError(
                "The number of new cyclic messages to be sent must be equal to "
                "the number of messages originally specified for this task"
            )
        if self.arbitration_id != messages[0].arbitration_id:
            raise ValueError(
                "The arbitration ID of new cyclic messages cannot be changed "
                "from when the task was created"
            )

    def modify_data(self, messages: Union[Sequence[Message], Message]) -> None:
        """Update the contents of the periodically sent messages, without
        altering the timing.

        :param messages:
            The messages with the new :attr:`Message.data`.

            Note: The arbitration ID cannot be changed.

            Note: The number of new cyclic messages to be sent must be equal
            to the original number of messages originally specified for this
            task.

        :raises ValueError: If the given messages are invalid
        """
        messages = self._check_and_convert_messages(messages)
        self._check_modified_messages(messages)

        self.messages = messages


class MultiRateCyclicSendTaskABC(CyclicSendTaskABC):
    """A Cyclic send task that supports switches send frequency after a set time."""

    def __init__(
        self,
        channel: typechecking.Channel,
        messages: Union[Sequence[Message], Message],
        count: int,  # pylint: disable=unused-argument
        initial_period: float,  # pylint: disable=unused-argument
        subsequent_period: float,
    ) -> None:
        """
        Transmits a message `count` times at `initial_period` then continues to
        transmit messages at `subsequent_period`.

        :param channel: See interface specific documentation.
        :param messages:
        :param count:
        :param initial_period:
        :param subsequent_period:

        :raises ValueError: If the given messages are invalid
        """
        super().__init__(messages, subsequent_period)
        self._channel = channel


class ThreadBasedCyclicSendTask(
    ModifiableCyclicTaskABC, LimitedDurationCyclicSendTaskABC, RestartableCyclicTaskABC
):
    """Fallback cyclic send task using daemon thread."""

    def __init__(
        self,
        bus: "BusABC",
        lock: threading.Lock,
        messages: Union[Sequence[Message], Message],
        period: float,
        duration: Optional[float] = None,
        on_error: Optional[Callable[[Exception], bool]] = None,
    ) -> None:
        """Transmits `messages` with a `period` seconds for `duration` seconds on a `bus`.

        The `on_error` is called if any error happens on `bus` while sending `messages`.
        If `on_error` present, and returns ``False`` when invoked, thread is
        stopped immediately, otherwise, thread continuously tries to send `messages`
        ignoring errors on a `bus`. Absence of `on_error` means that thread exits immediately
        on error.

        :param on_error: The callable that accepts an exception if any
                         error happened on a `bus` while sending `messages`,
                         it shall return either ``True`` or ``False`` depending
                         on desired behaviour of `ThreadBasedCyclicSendTask`.

        :raises ValueError: If the given messages are invalid
        """
        super().__init__(messages, period, duration)
        self.bus = bus
        self.send_lock = lock
        self.stopped = True
        self.thread: Optional[threading.Thread] = None
        self.end_time: Optional[float] = (
            time.perf_counter() + duration if duration else None
        )
        self.on_error = on_error

        if USE_WINDOWS_EVENTS:
            self.period_ms = int(round(period * 1000, 0))
            try:
                self.event = win32event.CreateWaitableTimerEx(
                    None,
                    None,
                    win32event.CREATE_WAITABLE_TIMER_HIGH_RESOLUTION,
                    win32event.TIMER_ALL_ACCESS,
                )
            except (AttributeError, OSError):
                self.event = win32event.CreateWaitableTimer(None, False, None)

        self.start()

    def stop(self) -> None:
        if USE_WINDOWS_EVENTS:
            win32event.CancelWaitableTimer(self.event.handle)
        self.stopped = True

    def start(self) -> None:
        self.stopped = False
        if self.thread is None or not self.thread.is_alive():
            name = f"Cyclic send task for 0x{self.messages[0].arbitration_id:X}"
            self.thread = threading.Thread(target=self._run, name=name)
            self.thread.daemon = True

            if USE_WINDOWS_EVENTS:
                win32event.SetWaitableTimer(
                    self.event.handle, 0, self.period_ms, None, None, False
                )

            self.thread.start()

    def _run(self) -> None:
        msg_index = 0
        msg_due_time_ns = time.perf_counter_ns()

        while not self.stopped:
            # Prevent calling bus.send from multiple threads
            with self.send_lock:
                try:
                    self.bus.send(self.messages[msg_index])
                except Exception as exc:  # pylint: disable=broad-except
                    log.exception(exc)
                    if self.on_error:
                        if not self.on_error(exc):
                            break
                    else:
                        break
            msg_due_time_ns += self.period_ns
            if self.end_time is not None and time.perf_counter() >= self.end_time:
                break
            msg_index = (msg_index + 1) % len(self.messages)

            # Compensate for the time it takes to send the message
            delay_ns = msg_due_time_ns - time.perf_counter_ns()

            if delay_ns > 0:
                if USE_WINDOWS_EVENTS:
                    win32event.WaitForSingleObject(
                        self.event.handle,
                        int(round(delay_ns / NANOSECONDS_IN_MILLISECOND)),
                    )
                else:
                    time.sleep(delay_ns / NANOSECONDS_IN_SECOND)



#!/usr/bin/env python

"""
Tests for CANtact interfaces
"""

import unittest

import can
from can.interfaces import cantact


class CantactTest(unittest.TestCase):
    def test_bus_creation(self):
        bus = can.Bus(channel=0, interface="cantact", _testing=True)
        self.assertIsInstance(bus, cantact.CantactBus)
        cantact.MockInterface.set_bitrate.assert_called()
        cantact.MockInterface.set_bit_timing.assert_not_called()
        cantact.MockInterface.set_enabled.assert_called()
        cantact.MockInterface.set_monitor.assert_called()
        cantact.MockInterface.start.assert_called()

    def test_bus_creation_bittiming(self):
        cantact.MockInterface.set_bitrate.reset_mock()

        bt = can.BitTiming(f_clock=24_000_000, brp=3, tseg1=13, tseg2=2, sjw=1)
        bus = can.Bus(channel=0, interface="cantact", timing=bt, _testing=True)
        self.assertIsInstance(bus, cantact.CantactBus)
        cantact.MockInterface.set_bitrate.assert_not_called()
        cantact.MockInterface.set_bit_timing.assert_called()
        cantact.MockInterface.set_enabled.assert_called()
        cantact.MockInterface.set_monitor.assert_called()
        cantact.MockInterface.start.assert_called()

    def test_transmit(self):
        bus = can.Bus(channel=0, interface="cantact", _testing=True)
        msg = can.Message(
            arbitration_id=0xC0FFEF, data=[1, 2, 3, 4, 5, 6, 7, 8], is_extended_id=True
        )
        bus.send(msg)
        cantact.MockInterface.send.assert_called()

    def test_recv(self):
        bus = can.Bus(channel=0, interface="cantact", _testing=True)
        frame = bus.recv(timeout=0.5)
        cantact.MockInterface.recv.assert_called()
        self.assertIsInstance(frame, can.Message)

    def test_recv_timeout(self):
        bus = can.Bus(channel=0, interface="cantact", _testing=True)
        frame = bus.recv(timeout=0.0)
        cantact.MockInterface.recv.assert_called()
        self.assertIsNone(frame)

    def test_shutdown(self):
        bus = can.Bus(channel=0, interface="cantact", _testing=True)
        bus.shutdown()
        cantact.MockInterface.stop.assert_called()



import importlib
from unittest.mock import patch

import pytest

import can
from can.interfaces import BACKENDS


@pytest.fixture(params=(BACKENDS.keys()))
def constructor(request):
    mod, cls = BACKENDS[request.param]

    try:
        module = importlib.import_module(mod)
        constructor = getattr(module, cls)
    except:
        pytest.skip("Unable to load interface")

    return constructor


@pytest.fixture
def interface(constructor):
    class MockInterface(constructor):
        def __init__(self):
            pass

        def __del__(self):
            pass

    return MockInterface()


@patch.object(can.bus.BusABC, "shutdown")
def test_interface_calls_parent_shutdown(mock_shutdown, interface):
    try:
        interface.shutdown()
    except:
        pass
    finally:
        mock_shutdown.assert_called()



#!/usr/bin/env python

import pickle
import sys
import unittest
from copy import copy, deepcopy
from datetime import timedelta
from math import isinf, isnan

import hypothesis.errors
import hypothesis.strategies as st
import pytest
from hypothesis import HealthCheck, given, settings

from can import Message

from .config import IS_GITHUB_ACTIONS, IS_PYPY, IS_WINDOWS
from .message_helper import ComparingMessagesTestCase


class TestMessageClass(unittest.TestCase):
    """
    This test tries many inputs to the message class constructor and then sanity checks
    all methods and ensures that nothing crashes. It also checks whether Message._check()
    allows all valid can frames.
    """

    @given(
        timestamp=st.floats(min_value=0.0),
        arbitration_id=st.integers(),
        is_extended_id=st.booleans(),
        is_remote_frame=st.booleans(),
        is_error_frame=st.booleans(),
        channel=st.one_of(st.text(), st.integers()),
        dlc=st.integers(min_value=0, max_value=8),
        data=st.one_of(st.binary(min_size=0, max_size=8), st.none()),
        is_fd=st.booleans(),
        bitrate_switch=st.booleans(),
        error_state_indicator=st.booleans(),
    )
    # The first run may take a second on CI runners and will hit the deadline
    @settings(
        max_examples=2000,
        suppress_health_check=[HealthCheck.too_slow],
        deadline=None if IS_GITHUB_ACTIONS else timedelta(milliseconds=500),
    )
    @pytest.mark.xfail(
        IS_WINDOWS and IS_PYPY,
        raises=hypothesis.errors.Flaky,
        reason="Hypothesis generates inconsistent timestamp floats on Windows+PyPy-3.7",
    )
    def test_methods(self, **kwargs):
        is_valid = not (
            (
                not kwargs["is_remote_frame"]
                and (len(kwargs["data"] or []) != kwargs["dlc"])
            )
            or (kwargs["arbitration_id"] >= 0x800 and not kwargs["is_extended_id"])
            or kwargs["arbitration_id"] >= 0x20000000
            or kwargs["arbitration_id"] < 0
            or (
                kwargs["is_remote_frame"]
                and (kwargs["is_fd"] or kwargs["is_error_frame"])
            )
            or (kwargs["is_remote_frame"] and len(kwargs["data"] or []) > 0)
            or (
                (kwargs["bitrate_switch"] or kwargs["error_state_indicator"])
                and not kwargs["is_fd"]
            )
            or isnan(kwargs["timestamp"])
            or isinf(kwargs["timestamp"])
        )

        # this should return normally and not throw an exception
        message = Message(check=is_valid, **kwargs)

        if kwargs["data"] is None or kwargs["is_remote_frame"]:
            kwargs["data"] = bytearray()

        if not is_valid and not kwargs["is_remote_frame"]:
            with self.assertRaises(ValueError):
                Message(check=True, **kwargs)

        self.assertGreater(len(str(message)), 0)
        self.assertGreater(len(message.__repr__()), 0)
        if is_valid:
            self.assertEqual(len(message), kwargs["dlc"])
        self.assertTrue(bool(message))
        self.assertGreater(len(f"{message}"), 0)
        _ = f"{message}"
        with self.assertRaises(Exception):
            _ = "{somespec}".format(
                message
            )  # pylint: disable=missing-format-argument-key
        if sys.version_info.major > 2:
            self.assertEqual(bytearray(bytes(message)), kwargs["data"])

        # check copies and equalities
        if is_valid:
            self.assertEqual(message, message)
            normal_copy = copy(message)
            deep_copy = deepcopy(message)
            for other in (normal_copy, deep_copy, message):
                self.assertTrue(message.equals(other, timestamp_delta=None))
                self.assertTrue(message.equals(other))
                self.assertTrue(message.equals(other, timestamp_delta=0))


class MessageSerialization(unittest.TestCase, ComparingMessagesTestCase):
    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        ComparingMessagesTestCase.__init__(
            self, allowed_timestamp_delta=0.016, preserves_channel=True
        )

    def test_serialization(self):
        message = Message(
            timestamp=1.0,
            arbitration_id=0x401,
            is_extended_id=False,
            is_remote_frame=False,
            is_error_frame=False,
            channel=1,
            dlc=6,
            data=bytearray([0x01, 0x02, 0x03, 0x04, 0x05, 0x06]),
            is_fd=False,
        )

        serialized = pickle.dumps(message, -1)
        deserialized = pickle.loads(serialized)

        self.assertMessageEqual(message, deserialized)


if __name__ == "__main__":
    unittest.main()



"""
Ctypes wrapper module for IXXAT Virtual CAN Interface V4 on win32 systems

Copyright (C) 2016 Giuseppe Corbelli <giuseppe.corbelli@weightpack.com>
"""

from . import structures

FALSE = 0
TRUE = 1

INFINITE = 0xFFFFFFFF

VCI_MAX_ERRSTRLEN = 256

# Bitrates
CAN_BT0_10KB = 0x31
CAN_BT1_10KB = 0x1C
CAN_BT0_20KB = 0x18
CAN_BT1_20KB = 0x1C
CAN_BT0_50KB = 0x09
CAN_BT1_50KB = 0x1C
CAN_BT0_100KB = 0x04
CAN_BT1_100KB = 0x1C
CAN_BT0_125KB = 0x03
CAN_BT1_125KB = 0x1C
CAN_BT0_250KB = 0x01
CAN_BT1_250KB = 0x1C
CAN_BT0_500KB = 0x00
CAN_BT1_500KB = 0x1C
CAN_BT0_667KB = 0x00
CAN_BT1_667KB = 0x18
CAN_BT0_800KB = 0x00
CAN_BT1_800KB = 0x16
CAN_BT0_1000KB = 0x00
CAN_BT1_1000KB = 0x14

# Facilities/severities
SEV_INFO = 0x40000000
SEV_WARN = 0x80000000
SEV_ERROR = 0xC0000000
SEV_MASK = 0xC0000000
SEV_SUCCESS = 0x00000000

RESERVED_FLAG = 0x10000000
CUSTOMER_FLAG = 0x20000000

STATUS_MASK = 0x0000FFFF
FACILITY_MASK = 0x0FFF0000

# Or so I hope
FACILITY_STD = 0

SEV_STD_INFO = SEV_INFO | CUSTOMER_FLAG | FACILITY_STD
SEV_STD_WARN = SEV_WARN | CUSTOMER_FLAG | FACILITY_STD
SEV_STD_ERROR = SEV_ERROR | CUSTOMER_FLAG | FACILITY_STD

FACILITY_VCI = 0x00010000
SEV_VCI_INFO = SEV_INFO | CUSTOMER_FLAG | FACILITY_VCI
SEV_VCI_WARN = SEV_WARN | CUSTOMER_FLAG | FACILITY_VCI
SEV_VCI_ERROR = SEV_ERROR | CUSTOMER_FLAG | FACILITY_VCI

FACILITY_DAL = 0x00020000
SEV_DAL_INFO = SEV_INFO | CUSTOMER_FLAG | FACILITY_DAL
SEV_DAL_WARN = SEV_WARN | CUSTOMER_FLAG | FACILITY_DAL
SEV_DAL_ERROR = SEV_ERROR | CUSTOMER_FLAG | FACILITY_DAL

FACILITY_CCL = 0x00030000
SEV_CCL_INFO = SEV_INFO | CUSTOMER_FLAG | FACILITY_CCL
SEV_CCL_WARN = SEV_WARN | CUSTOMER_FLAG | FACILITY_CCL
SEV_CCL_ERROR = SEV_ERROR | CUSTOMER_FLAG | FACILITY_CCL

FACILITY_BAL = 0x00040000
SEV_BAL_INFO = SEV_INFO | CUSTOMER_FLAG | FACILITY_BAL
SEV_BAL_WARN = SEV_WARN | CUSTOMER_FLAG | FACILITY_BAL
SEV_BAL_ERROR = SEV_ERROR | CUSTOMER_FLAG | FACILITY_BAL

# Errors
VCI_SUCCESS = 0x00
VCI_OK = 0x00
VCI_E_UNEXPECTED = SEV_VCI_ERROR | 0x0001
VCI_E_NOT_IMPLEMENTED = SEV_VCI_ERROR | 0x0002
VCI_E_OUTOFMEMORY = SEV_VCI_ERROR | 0x0003
VCI_E_INVALIDARG = SEV_VCI_ERROR | 0x0004
VCI_E_NOINTERFACE = SEV_VCI_ERROR | 0x0005
VCI_E_INVPOINTER = SEV_VCI_ERROR | 0x0006
VCI_E_INVHANDLE = SEV_VCI_ERROR | 0x0007
VCI_E_ABORT = SEV_VCI_ERROR | 0x0008
VCI_E_FAIL = SEV_VCI_ERROR | 0x0009
VCI_E_ACCESSDENIED = SEV_VCI_ERROR | 0x000A
VCI_E_TIMEOUT = SEV_VCI_ERROR | 0x000B
VCI_E_BUSY = SEV_VCI_ERROR | 0x000C
VCI_E_PENDING = SEV_VCI_ERROR | 0x000D
VCI_E_NO_DATA = SEV_VCI_ERROR | 0x000E
VCI_E_NO_MORE_ITEMS = SEV_VCI_ERROR | 0x000F
VCI_E_NOT_INITIALIZED = SEV_VCI_ERROR | 0x0010
VCI_E_ALREADY_INITIALIZED = SEV_VCI_ERROR | 0x00011
VCI_E_RXQUEUE_EMPTY = SEV_VCI_ERROR | 0x00012
VCI_E_TXQUEUE_FULL = SEV_VCI_ERROR | 0x0013
VCI_E_BUFFER_OVERFLOW = SEV_VCI_ERROR | 0x0014
VCI_E_INVALID_STATE = SEV_VCI_ERROR | 0x0015
VCI_E_OBJECT_ALREADY_EXISTS = SEV_VCI_ERROR | 0x0016
VCI_E_INVALID_INDEX = SEV_VCI_ERROR | 0x0017
VCI_E_END_OF_FILE = SEV_VCI_ERROR | 0x0018
VCI_E_DISCONNECTED = SEV_VCI_ERROR | 0x0019
VCI_E_WRONG_FLASHFWVERSION = SEV_VCI_ERROR | 0x001A

# Controller status
CAN_STATUS_TXPEND = 0x01  # transmission pending
CAN_STATUS_OVRRUN = 0x02  # data overrun occurred
CAN_STATUS_ERRLIM = 0x04  # error warning limit exceeded
CAN_STATUS_BUSOFF = 0x08  # bus off status
CAN_STATUS_ININIT = 0x10  # init mode active
CAN_STATUS_BUSCERR = 0x20  # bus coupling error

# Controller operating modes
CAN_OPMODE_UNDEFINED = 0x00  # undefined
CAN_OPMODE_STANDARD = 0x01  # reception of 11-bit id messages
CAN_OPMODE_EXTENDED = 0x02  # reception of 29-bit id messages
CAN_OPMODE_ERRFRAME = 0x04  # reception of error frames
CAN_OPMODE_LISTONLY = 0x08  # listen only mode (TX passive)
CAN_OPMODE_LOWSPEED = 0x10  # use low speed bus interface
CAN_OPMODE_AUTOBAUD = 0x20  # automatic bit rate detection

# Extended operating modes
CAN_EXMODE_DISABLED = 0x00
CAN_EXMODE_EXTDATALEN = 0x01
CAN_EXMODE_FASTDATA = 0x02
CAN_EXMODE_NONISOCANFD = 0x04

# Message types
CAN_MSGTYPE_DATA = 0
CAN_MSGTYPE_INFO = 1
CAN_MSGTYPE_ERROR = 2
CAN_MSGTYPE_STATUS = 3
CAN_MSGTYPE_WAKEUP = 4
CAN_MSGTYPE_TIMEOVR = 5
CAN_MSGTYPE_TIMERST = 6

# Information supplied in the abData[0] field of info frames
# (CANMSGINFO.Bytes.bType = CAN_MSGTYPE_INFO).
CAN_INFO_START = 1
CAN_INFO_STOP = 2
CAN_INFO_RESET = 3

# Information supplied in the abData[0] field of info frames
# (CANMSGINFO.Bytes.bType = CAN_MSGTYPE_ERROR).
CAN_ERROR_STUFF = 1  # stuff error
CAN_ERROR_FORM = 2  # form error
CAN_ERROR_ACK = 3  # acknowledgment error
CAN_ERROR_BIT = 4  # bit error
CAN_ERROR_CRC = 6  # CRC error
CAN_ERROR_OTHER = 7  # other (unspecified) error

# acceptance code and mask to reject all CAN IDs
CAN_ACC_MASK_NONE = 0xFFFFFFFF
CAN_ACC_CODE_NONE = 0x80000000

# BTMODEs
CAN_BTMODE_RAW = 0x00000001  # raw mode
CAN_BTMODE_TSM = 0x00000002  # triple sampling mode


CAN_FILTER_VOID = 0x00  # invalid or unknown filter mode (do not use for initialization)
CAN_FILTER_LOCK = 0x01  # lock filter (inhibit all IDs)
CAN_FILTER_PASS = 0x02  # bypass filter (pass all IDs)
CAN_FILTER_INCL = 0x03  # inclusive filtering (pass registered IDs)
CAN_FILTER_EXCL = 0x04  # exclusive filtering (inhibit registered IDs)


# message information flags (used by <CANMSGINFO.Bytes.bFlags>)
CAN_MSGFLAGS_DLC = 0x0F  # [bit 0] data length code
CAN_MSGFLAGS_OVR = 0x10  # [bit 4] data overrun flag
CAN_MSGFLAGS_SRR = 0x20  # [bit 5] self reception request
CAN_MSGFLAGS_RTR = 0x40  # [bit 6] remote transmission request
CAN_MSGFLAGS_EXT = 0x80  # [bit 7] frame format (0=11-bit, 1=29-bit)

# extended message information flags (used by <CANMSGINFO.Bytes.[bFlags2|bAddFlags]>)
CAN_MSGFLAGS2_SSM = 0x01  # [bit 0] single shot mode
CAN_MSGFLAGS2_HPM = 0x02  # [bit 1] high priority message
CAN_MSGFLAGS2_EDL = 0x04  # [bit 2] extended data length
CAN_MSGFLAGS2_FDR = 0x08  # [bit 3] fast data bit rate
CAN_MSGFLAGS2_ESI = 0x10  # [bit 4] error state indicator
CAN_MSGFLAGS2_RES = 0xE0  # [bit 5..7] reserved bits


CAN_ACCEPT_REJECT = 0x00  # message not accepted
CAN_ACCEPT_ALWAYS = 0xFF  # message always accepted
CAN_ACCEPT_FILTER_1 = 0x01  # message accepted by filter 1
CAN_ACCEPT_FILTER_2 = 0x02  # message accepted by filter 2
CAN_ACCEPT_PASSEXCL = 0x03  # message passes exclusion filter


CAN_FEATURE_STDOREXT = 0x00000001  # 11 OR 29 bit (exclusive)
CAN_FEATURE_STDANDEXT = 0x00000002  # 11 AND 29 bit (simultaneous)
CAN_FEATURE_RMTFRAME = 0x00000004  # reception of remote frames
CAN_FEATURE_ERRFRAME = 0x00000008  # reception of error frames
CAN_FEATURE_BUSLOAD = 0x00000010  # bus load measurement
CAN_FEATURE_IDFILTER = 0x00000020  # exact message filter
CAN_FEATURE_LISTONLY = 0x00000040  # listen only mode
CAN_FEATURE_SCHEDULER = 0x00000080  # cyclic message scheduler
CAN_FEATURE_GENERRFRM = 0x00000100  # error frame generation
CAN_FEATURE_DELAYEDTX = 0x00000200  # delayed message transmitter
CAN_FEATURE_SINGLESHOT = 0x00000400  # single shot mode
CAN_FEATURE_HIGHPRIOR = 0x00000800  # high priority message
CAN_FEATURE_AUTOBAUD = 0x00001000  # automatic bit rate detection
CAN_FEATURE_EXTDATA = 0x00002000  # extended data length (CANFD)
CAN_FEATURE_FASTDATA = 0x00004000  # fast data bit rate (CANFD)
CAN_FEATURE_ISOFRAME = 0x00008000  # ISO conform frame (CANFD)
CAN_FEATURE_NONISOFRM = (
    0x00010000  # non ISO conform frame (CANFD) (different CRC computation)
)
CAN_FEATURE_64BITTSC = 0x00020000  # 64-bit time stamp counter


CAN_BITRATE_PRESETS = {
    250000: structures.CANBTP(
        dwMode=0, dwBPS=250000, wTS1=6400, wTS2=1600, wSJW=1600, wTDO=0
    ),  # SP = 80,0%
    500000: structures.CANBTP(
        dwMode=0, dwBPS=500000, wTS1=6400, wTS2=1600, wSJW=1600, wTDO=0
    ),  # SP = 80,0%
    1000000: structures.CANBTP(
        dwMode=0, dwBPS=1000000, wTS1=6400, wTS2=1600, wSJW=1600, wTDO=0
    ),  # SP = 80,0%
}

CAN_DATABITRATE_PRESETS = {
    500000: structures.CANBTP(
        dwMode=0, dwBPS=500000, wTS1=6400, wTS2=1600, wSJW=1600, wTDO=6400
    ),  # SP = 80,0%
    833333: structures.CANBTP(
        dwMode=0, dwBPS=833333, wTS1=1600, wTS2=400, wSJW=400, wTDO=1620
    ),  # SP = 80,0%
    1000000: structures.CANBTP(
        dwMode=0, dwBPS=1000000, wTS1=1600, wTS2=400, wSJW=400, wTDO=1600
    ),  # SP = 80,0%
    1538461: structures.CANBTP(
        dwMode=0, dwBPS=1538461, wTS1=1000, wTS2=300, wSJW=300, wTDO=1040
    ),  # SP = 76,9%
    2000000: structures.CANBTP(
        dwMode=0, dwBPS=2000000, wTS1=1600, wTS2=400, wSJW=400, wTDO=1600
    ),  # SP = 80,0%
    4000000: structures.CANBTP(
        dwMode=0, dwBPS=4000000, wTS1=800, wTS2=200, wSJW=200, wTDO=800
    ),  # SP = 80,0%
    5000000: structures.CANBTP(
        dwMode=0, dwBPS=5000000, wTS1=600, wTS2=200, wSJW=200, wTDO=600
    ),  # SP = 75,0%
    6666666: structures.CANBTP(
        dwMode=0, dwBPS=6666666, wTS1=400, wTS2=200, wSJW=200, wTDO=402
    ),  # SP = 66,7%
    8000000: structures.CANBTP(
        dwMode=0, dwBPS=8000000, wTS1=400, wTS2=100, wSJW=100, wTDO=250
    ),  # SP = 80,0%
    10000000: structures.CANBTP(
        dwMode=0, dwBPS=10000000, wTS1=300, wTS2=100, wSJW=100, wTDO=200
    ),  # SP = 75,0%
}
